#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>

#include <gtk/gtk.h>

#include "gtklife.h"
#include "life.h"
#include "ewmh.h"
#include "util.h"
#include "icons.h"

#define SWAP(a,b) {temp=a; a=b; b=temp;}

/*** Constants ***/

static point  null_point = {-1, -1};
static rect   null_rect  = {{-1, -1}, {-1, -1}};

/* Menubar
 * Fields: Item, Keybinding, Handler function, (Unused), Item type
 */
static GtkItemFactoryEntry menu_items[] = {
    { "/_File",                  0,                 0,                      0,  "<Branch>"     },
    { "/File/tearoff",           0,                 0,                      0,  "<Tearoff>"    },
    { "/File/_New",              "<control>N",      file_new,               0,  0              },
    { "/File/_Open...",          "<control>O",      file_open,              0,  0              },
    { "/File/_Reopen",           "R",               file_reopen,            0,  0              },
    { "/File/_Save",             "<control>S",      file_save,              0,  0              },
    { "/File/Save _As...",       "<control>A",      file_save_as,           0,  0              },
    { "/File/sep1",              0,                 0,                      0,  "<Separator>"  },
    { "/File/_Description...",   "D",               file_description,       0,  0              },
    { "/File/sep2",              0,                 0,                      0,  "<Separator>"  },
    { "/File/_Quit",             "<control>Q",      file_quit,              0,  0              },

    { "/_View",                  0,                 0,                      0,  "<Branch>"     },
    { "/View/tearoff",           0,                 0,                      0,  "<Tearoff>"    },
    { "/View/Zoom _In",          "equal",           view_zoom_in,           0,  0              },
    { "/View/Zoom _Out",         "minus",           view_zoom_out,          0,  0              },
    { "/View/_Zoom",             0,                 0,                      0,  "<Branch>"     },
    { "/View/Zoom/tearoff",      0,                 0,                      0,  "<Tearoff>"    },
    { "/View/Zoom/_1:1",         "1",               view_zoom_1,            0,  0              },
    { "/View/Zoom/_2:1",         "2",               view_zoom_2,            0,  0              },
    { "/View/Zoom/_4:1",         "4",               view_zoom_4,            0,  0              },
    { "/View/Zoom/_8:1",         "8",               view_zoom_8,            0,  0              },
    { "/View/Zoom/1_6:1",        "Z",               view_zoom_16,           0,  0              },
    { "/View/sep1",              0,                 0,                      0,  "<Separator>"  },
    { "/View/_Recenter",         "KP_5",            view_recenter,          0,  0              },
    { "/View/sep2",              0,                 0,                      0,  "<Separator>"  },
    { "/View/_Fullscreen",       "F",               view_fullscreen,        0,  "<CheckItem>"  },
    { "/View/Show _Toolbar",     0,                 view_show_toolbar,      0,  "<CheckItem>"  },
    { "/View/Show _Sidebar",     0,                 view_show_sidebar,      0,  "<CheckItem>"  },
    { "/View/Show S_crollbars",  0,                 view_show_scrollbars,   0,  "<CheckItem>"  },

    /* Invisible submenu, just used to set accelerators */
    { "/Position",               0,                 0,                      0,  "<Branch>"     },
    { "/Position/Scroll Left",   "KP_4",            position_scroll_left,   0,  0              },
    { "/Position/Scroll Right",  "KP_6",            position_scroll_right,  0,  0              },
    { "/Position/Scroll Up",     "KP_8",            position_scroll_up,     0,  0              },
    { "/Position/Scroll Down",   "KP_2",            position_scroll_down,   0,  0              },
    { "/Position/Scroll NW",     "KP_7",            position_scroll_nw,     0,  0              },
    { "/Position/Scroll NE",     "KP_9",            position_scroll_ne,     0,  0              },
    { "/Position/Scroll SW",     "KP_1",            position_scroll_sw,     0,  0              },
    { "/Position/Scroll SE",     "KP_3",            position_scroll_se,     0,  0              },
    { "/Position/Page Left",     "<control>KP_4",   position_page_left,     0,  0              },
    { "/Position/Page Right",    "<control>KP_6",   position_page_right,    0,  0              },
    { "/Position/Page Up",       "<control>KP_8",   position_page_up,       0,  0              },
    { "/Position/Page Down",     "<control>KP_2",   position_page_down,     0,  0              },
    { "/Position/Page NW",       "<control>KP_7",   position_page_nw,       0,  0              },
    { "/Position/Page NE",       "<control>KP_9",   position_page_ne,       0,  0              },
    { "/Position/Page SW",       "<control>KP_1",   position_page_sw,       0,  0              },
    { "/Position/Page SE",       "<control>KP_3",   position_page_se,       0,  0              },

    { "/_Edit",                  0,                 0,                      0,  "<Branch>"     },
    { "/Edit/tearoff",           0,                 0,                      0,  "<Tearoff>"    },
    { "/Edit/_Copy",             "<control>C",      edit_copy,              0,  0              },
    { "/Edit/C_ut",              "<control>X",      edit_cut,               0,  0              },
    { "/Edit/_Paste",            "<control>V",      edit_paste,             0,  0              },
    { "/Edit/C_lear",            "<control>D",      edit_clear,             0,  0              },
    { "/Edit/_Move",             "<alt>M",          edit_move,              0,  0              },
    { "/Edit/sep1",              0,                 0,                      0,  "<Separator>"  },
    { "/Edit/Pre_ferences",      "P",               edit_preferences,       0,  0              },

    { "/_Run",                   0,                 0,                      0,  "<Branch>"     },
    { "/Run/tearoff",            0,                 0,                      0,  "<Tearoff>"    },
    { "/Run/_Start",             "S",               run_start_stop,         0,  0              },
    { "/Run/S_tep",              "T",               run_step,               0,  0              },
    { "/Run/sep1",               0,                 0,                      0,  "<Separator>"  },
    { "/Run/_Faster",            "period",          run_faster,             0,  0              },
    { "/Run/Slo_wer",            "comma",           run_slower,             0,  0              },
    { "/Run/S_peed...",          "<alt>S",          run_speed,              0,  0              },

    { "/_Help",                  0,                 0,                      0,  "<LastBranch>" },
    { "/Help/tearoff",           0,                 0,                      0,  "<Tearoff>"    },
    { "/Help/_Help...",          "F1",              help_help,              0,  0              },
    { "/Help/_About...",         0,                 help_about,             0,  0              }
};

/* Toolbar
 * Fields: Tooltip, Icon data, Handler function
 */
static toolbar_item toolbar_items[] = {
    { "Clear the grid",                 new_icon,          file_new         },
    { "Open a pattern file",            open_icon,         file_open        },
    { "Save pattern",                   save_icon,         file_save        },
    { "Save pattern as",                save_as_icon,      file_save_as     },
    { "View/edit pattern description",  description_icon,  file_description },

    {"sep",NULL,NULL},

    { "Zoom in",                        zoom_in_icon,      view_zoom_in     },
    { "Zoom out",                       zoom_out_icon,     view_zoom_out    },
    { "Zoom 1:1",                       zoom_1_icon,       view_zoom_1      },
    { "Zoom all the way in",            zoom_16_icon,      view_zoom_16     },

    {"sep",NULL,NULL},

    { "Edit preferences",               preferences_icon,  edit_preferences },

    {"sep",NULL,NULL},

    { "Start/stop",                     start_icon,        run_start_stop   },
    { "Step ahead one generation",      step_icon,         run_step         },

    {NULL,NULL,NULL}
};

/*** Globals ***/

static char*  user_dir = NULL;

/* Preferences */
static struct {
    char*     patterns_dir;
    int32     default_window_width;
    int32     default_window_height;
    int32     default_zoom;
    boolean   default_show_toolbar;
    boolean   default_show_sidebar;
    boolean   default_show_scrollbars;
    boolean   default_fullscreen;
    uint32    colors[NUM_COLORS];
    int32     default_speed;
    int32     speed_slider_min;
    int32     speed_slider_max;
    int32     speed_increment;
    boolean   default_skip_frames;
    char*     help_browser;
} config = {
    DEFAULT_PATTERNS_DIR,
    DEFAULT_WINDOW_WIDTH,
    DEFAULT_WINDOW_HEIGHT,
    DEFAULT_ZOOM,
    DEFAULT_SHOW_TOOLBAR,
    DEFAULT_SHOW_SIDEBAR,
    DEFAULT_SHOW_SCROLLBARS,
    DEFAULT_FULLSCREEN,
    {DEFAULT_BG_COLOR, DEFAULT_CELL_COLOR, DEFAULT_GRID_COLOR, DEFAULT_SELECT_COLOR},
    DEFAULT_SPEED,
    DEFAULT_SPEED_SLIDER_MIN,
    DEFAULT_SPEED_SLIDER_MAX,
    DEFAULT_SPEED_INCREMENT,
    DEFAULT_SKIP_FRAMES,
    DEFAULT_HELP_BROWSER
};

/* Some general state globals */
static struct {
    /* Directories and pattern paths */
    char*        current_dir;
    char*        pattern_path;
    recent_file  recent_files[MAX_RECENT_FILES];

    /* Running pattern stats */
    boolean  pattern_running;
    int32    speed;
    boolean  skip_frames;
    uint32   start_tick;
    uint64   start_time;
    int32    skipped_frames;

    /* Data related to mouse dragging and cut/paste */
    tracking_mouse_type  tracking_mouse;
    point       last_drawn;
    rect        selection;
    uint64      selection_start_time;
    cage_type*  copy_buffer;
    rect        copy_rect;
    rect        paste_box;
    boolean     moving;
} state;

/* Globals related to the GUI display state */
static struct {
    /* Zoom level, where [zoom]:1 is the display ratio */
    int32        zoom;

    /* Physical canvas and logical viewport. Canvas is measured in pixels,
     * viewport in cells */
    dimension    canvas_size;
    dimension    eff_canvas_size;   /* "effective" canvas size--minus any unused space */
    dimension    viewport_sizes[MAX_ZOOM+1];
    dimension    viewport_size;
    rect         viewport;

    /* Clipping rectangle, in pixels */
    rect         update;

    /* Offscreen pixmap data */
    uint8*       life_pixmap;
    uint32       life_pixmap_alloc_len;
    GdkRgbCmap*  life_pixmap_colormap;

    /* Which parts of the GUI are visible */
    boolean      show_toolbar;
    boolean      show_sidebar;
    boolean      show_scrollbars;
    boolean      fullscreen;
} dstate;

/* GUI objects that need to be global */
static struct {
    GtkWidget*   window;
    GtkWidget*   menubar;
    GtkWidget*   toolbar_box;
    GtkWidget*   sidebar;
    GtkWidget*   pattern_list;
    GtkWidget*   canvas;
    GtkWidget*   vscrollbar;
    GtkWidget*   hscrollbar;
    GtkWidget*   recent_files_separator;
    GtkWidget*   show_toolbar_menu_item;
    GtkWidget*   show_sidebar_menu_item;
    GtkWidget*   show_scrollbars_menu_item;
    GtkWidget*   position_submenu;
    GtkWidget*   start_stop_menu_label;
    GtkWidget*   start_stop_toolbar_button;
    GtkWidget*   start_pixmap;
    GtkWidget*   stop_pixmap;
    GtkWidget*   speed_label;
    GtkWidget*   speed_slider;
    GtkWidget*   tick_label;
    GtkWidget*   population_label;
} gui;

/*** Main ***/

int main(int argc, char *argv[]) {
    set_prog_name(PROG);
    create_user_dir();
    load_preferences();
    apply_defaults();
    load_recent_files_list();
    gtk_init(&argc, &argv);
    init_gui();
    load_keybindings();
    if (argc > 1)
        attempt_load_pattern(argv[1]);
    main_loop();

    return 0;
}

/*** Functions for reading/saving user settings ***/

/* Create the user directory (if necessary), and record its full path.
 */
void create_user_dir(void)
{
    const char*  home_dir;

    home_dir = getenv("HOME");
    if (!home_dir) {
        warn("No home directory found! Setting user dir to '/'");
        user_dir = "";
        return;
    }
    user_dir = dsprintf("%s/%s", home_dir, USER_DIR);
    if (mkdir(user_dir, 0777) < 0 && errno != EEXIST) {
        sys_warn("can't create user directory %s", user_dir);
        return;
    }
}

/* Read and apply saved user preferences
 */
void load_preferences(void)
{
    FILE*  f;
    char*  path;
    char*  line;
    char*  key;
    char*  val;
    char*  equal_pos;

    config.patterns_dir = safe_strdup(config.patterns_dir);
    config.help_browser = safe_strdup(config.help_browser);

    path = dsprintf("%s/preferences", user_dir);
    f = fopen(path, "r");
    free(path);
    if (!f)
        return;

    while ((line = dgets(f)) != NULL) {
        equal_pos = strchr(line, '=');
        if (!equal_pos) {
            warn("invalid line '%s' found in preferences file, ignoring", line);
            free(line);
            continue;
        }
        *equal_pos = '\0';
        key = line;
        val = equal_pos+1;

        if (strstr(key, "patterns_dir")) {
            free(config.patterns_dir);
            config.patterns_dir = safe_strdup(val);
        } else if (strstr(key, "window_width"))
            config.default_window_width = atoi(val);
        else if (strstr(key, "window_height"))
            config.default_window_height = atoi(val);
        else if (STR_EQUAL(key, "default_zoom"))
            config.default_zoom = atoi(val);
        else if (STR_EQUAL(key, "default_show_toolbar"))
            config.default_show_toolbar = atoi(val);
        else if (STR_EQUAL(key, "default_show_sidebar"))
            config.default_show_sidebar = atoi(val);
        else if (STR_EQUAL(key, "default_show_scrollbars"))
            config.default_show_scrollbars = atoi(val);
        else if (STR_EQUAL(key, "default_fullscreen"))
            config.default_fullscreen = atoi(val);
        else if (STR_STARTS_WITH(key, "color_") && is_numeric(key+6) && atoi(key+6) < NUM_COLORS)
            config.colors[atoi(key+6)] = strtoul(val, NULL, 16);
        else if (STR_EQUAL(key, "default_speed"))
            config.default_speed = atoi(val);
        else if (STR_EQUAL(key, "speed_slider_min"))
            config.speed_slider_min = atoi(val);
        else if (STR_EQUAL(key, "speed_slider_max"))
            config.speed_slider_max = atoi(val);
        else if (STR_EQUAL(key, "speed_increment"))
            config.speed_increment = atoi(val);
        else if (STR_EQUAL(key, "default_skip_frames"))
            config.default_skip_frames = atoi(val);
        else if (STR_EQUAL(key, "help_browser")) {
            free(config.help_browser);
            config.help_browser = safe_strdup(val);
        } else
            warn("invalid key '%s' found in preferences file, ignoring", key);

        free(line);
    }

    fclose(f);
}

/* Apply default values to various globals, based on preferences and/or system defaults.
 */
void apply_defaults(void)
{
    state.current_dir      = safe_strdup(config.patterns_dir);
    state.speed            = config.default_speed;
    state.skip_frames      = config.default_skip_frames;
    dstate.zoom            = config.default_zoom;
    dstate.show_toolbar    = config.default_show_toolbar;
    dstate.show_sidebar    = config.default_show_sidebar;
    dstate.show_scrollbars = config.default_show_scrollbars;

    state.selection = state.paste_box = null_rect;
    state.last_drawn = null_point;
}

/* Read and apply saved user keybindings
 */
void load_keybindings(void)
{
    char*  path;

    path = dsprintf("%s/keybindings", user_dir);
    gtk_item_factory_parse_rc(path);
    free(path);
}

/* Read the recent file list to put in the "File" menu
 */
void load_recent_files_list(void)
{
    FILE*  f;
    char*  rf_path;
    char*  full_path;
    char*  filename;
    char*  slash_pos;
    int32  i;

    memset(state.recent_files, 0, sizeof(recent_file) * MAX_RECENT_FILES);
    rf_path = dsprintf("%s/recent", user_dir);
    f = fopen(rf_path, "r");
    free(rf_path);
    if (!f)
        return;
    i = 0;
    while (i < MAX_RECENT_FILES && (full_path = dgets(f)) != NULL) {
        slash_pos = strrchr(full_path, '/');
        filename = safe_strdup(slash_pos ? slash_pos+1 : full_path);
        state.recent_files[i].full_path = full_path;
        state.recent_files[i].filename  = filename;
        i++;
    }
    fclose(f);
}

/* Save current preferences to a file in the user directory.
 */
void save_preferences(void)
{
    FILE*  f;
    char*  path;
    int32  i;

    path = dsprintf("%s/preferences", user_dir);
    f = fopen(path, "w");
    free(path);
    if (!f) {
        sys_warn("can't open %s/preferences for writing", user_dir);
        return;
    }

    fprintf(f, "patterns_dir=%s\n",             config.patterns_dir);
    fprintf(f, "default_window_width=%d\n",     config.default_window_width);
    fprintf(f, "default_window_height=%d\n",    config.default_window_height);
    fprintf(f, "default_zoom=%d\n",             config.default_zoom);
    fprintf(f, "default_show_toolbar=%d\n",     config.default_show_toolbar);
    fprintf(f, "default_show_sidebar=%d\n",     config.default_show_sidebar);
    fprintf(f, "default_show_scrollbars=%d\n",  config.default_show_scrollbars);
    fprintf(f, "default_fullscreen=%d\n",       config.default_fullscreen);
    for (i=0; i < NUM_COLORS; i++)
        fprintf(f, "color_%d=%06X\n", i, config.colors[i]);
    fprintf(f, "default_speed=%d\n",            config.default_speed);
    fprintf(f, "speed_slider_min=%d\n",         config.speed_slider_min);
    fprintf(f, "speed_slider_max=%d\n",         config.speed_slider_max);
    fprintf(f, "speed_increment=%d\n",          config.speed_increment);
    fprintf(f, "default_skip_frames=%d\n",      config.default_skip_frames);
    fprintf(f, "help_browser=%s\n",             config.help_browser);

    if (fclose(f) != 0)
        sys_warn("close failed after writing to ~/%s/preferences", USER_DIR);
}

/* Save current keybindings to a file in the user directory.
 */
void save_keybindings(void)
{
    char*  path;

    path = dsprintf("%s/keybindings", user_dir);
    gtk_item_factory_dump_rc(path, NULL, TRUE);
    free(path);
}

/* Save the list of recently accessed pattern files
 */
void save_recent_files_list(void)
{
    FILE*  f;
    char*  path;
    int32  i;

    path = dsprintf("%s/recent", user_dir);
    f = fopen(path, "w");
    free(path);
    if (!f) {
        sys_warn("can't open %s/recent for writing", user_dir);
        return;
    }
    for (i=0; i < MAX_RECENT_FILES && state.recent_files[i].full_path; i++)
        fprintf(f, "%s\n", state.recent_files[i].full_path);
    if (fclose(f) != 0)
        sys_warn("close failed after writing to ~/%s/recent", USER_DIR);
}

/* Main loop functions */

/* Handle events and run the Life pattern
 */
void main_loop(void)
{
    boolean block;

    while (1) {
        /* Block if no pattern is running and we don't need to track mouse movement, else poll */
        while ((block = (!state.pattern_running && !state.tracking_mouse)) || gtk_events_pending())
            gtk_main_iteration_do(block);
        if (state.tracking_mouse)
            handle_mouse_movement();
        if (state.pattern_running)
            tick_and_update();
        else
            usleep(1);   /* don't eat up all the processor just for mouse dragging */
    }
}

/* If we're in sync with the requested speed (or the pattern is being stepped through manually),
 * process the next Life tick, update the pixmap and redraw the screen. If going too fast, don't do
 * anything. If too slow and state.skip_frames is TRUE, and state.skipped_frames is not maxed out
 * yet, process the next tick and update the pixmap, but don't redraw the screen.
 */
void tick_and_update(void)
{
    int32  running_fps = 0;

    if (state.pattern_running) {
        running_fps = ROUND((double)(tick - state.start_tick) * 1000.0 /
                            (double)(get_time_milliseconds() - state.start_time));
        if (running_fps > state.speed)
            return;
    }
    if (!state.skipped_frames) {
        dstate.update.start.x = dstate.canvas_size.width  - 1;
        dstate.update.start.y = dstate.canvas_size.height - 1;
        dstate.update.end.x = dstate.update.end.y = 0;
    }

    next_tick();

    if (state.pattern_running && running_fps < state.speed && state.skip_frames &&
        state.skipped_frames < MAX_SKIPPED_FRAMES)
        state.skipped_frames++;
    else {
        trigger_canvas_update();
        update_tick_label();
        update_population_label();
    }
}

/*** GUI initialization functions ***/

/* Setup the gtklife window and widgets
 */
void init_gui(void)
{
    GtkWidget*  vbox;
    GtkWidget*  hbox;
    int32       i;

    init_rgb();
    init_window();

    vbox = gtk_vbox_new(FALSE, 0);
    init_menubar(vbox);
    init_toolbar(vbox);
    hbox = gtk_hbox_new(FALSE, 0);
    init_sidebar(hbox);
    init_canvas(hbox);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
    gtk_container_add(GTK_CONTAINER(gui.window), vbox);
    gtk_widget_show_all(gui.window);

    /* Hide the "Position" submenu -- we only use it to set accelerators */
    gtk_widget_hide(gui.position_submenu);

    /* Hide any GUI components the user doesn't want to see */
    if (!dstate.show_toolbar)
        gtk_widget_hide_all(gui.toolbar_box);
    if (!dstate.show_sidebar)
        gtk_widget_hide_all(gui.sidebar);
    if (!dstate.show_scrollbars) {
        gtk_widget_hide(gui.vscrollbar);
        gtk_widget_hide(gui.hscrollbar);
    }

    /* Hide unused recent-file entries. If there are no entries, hide the separator too. */
    for (i=0; i < MAX_RECENT_FILES; i++) {
        if (!state.recent_files[i].filename)
            gtk_widget_hide(state.recent_files[i].menu_item);
    }
    if (!state.recent_files[0].filename)
        gtk_widget_hide(gui.recent_files_separator);

    /* Put keyboard focus on the speed slider */
    gtk_widget_grab_focus(gui.speed_slider);

    /* Go to fullscreen if config says to */
    if (config.default_fullscreen)
        view_fullscreen();
}

/* Initialize the GdkRGB subsystem for image handling. Also create the colormap we'll be using.
 */
void init_rgb(void)
{
    gdk_rgb_init();
    gtk_widget_set_default_colormap(gdk_rgb_get_cmap());
    gtk_widget_set_default_visual(gdk_rgb_get_visual());
    dstate.life_pixmap_colormap = gdk_rgb_cmap_new(config.colors, NUM_COLORS);
}

/* Initialize the main application window
 */
void init_window(void)
{
    gui.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect(GTK_OBJECT(gui.window), "destroy", GTK_SIGNAL_FUNC(file_quit),
                       "WM Destroy");
    gtk_widget_set_name(GTK_WIDGET(gui.window), PROG);
    gtk_window_set_title(GTK_WINDOW(gui.window), TITLE);
    gtk_window_set_default_size(GTK_WINDOW(gui.window), config.default_window_width,
                                config.default_window_height);
    gtk_widget_realize(gui.window);
}

/* Setup the top menubar
 */
void init_menubar(GtkWidget* containing_box)
{
    GtkItemFactory*  menu_factory;
    GtkAccelGroup*   accelerators;
    GtkWidget*       file_menu;
    GtkWidget*       start_stop_menu_item;
    GtkWidget*       item;
    char*            str;
    int32            num_items;
    int32*           num;
    int32            i;

    /* Setup menu items */
    accelerators = gtk_accel_group_new();
    menu_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accelerators);
    num_items = sizeof(menu_items) / sizeof(menu_items[0]);
    gtk_item_factory_create_items(menu_factory, num_items, menu_items, NULL);
    gtk_window_add_accel_group(GTK_WINDOW(gui.window), accelerators);
    gui.menubar = gtk_item_factory_get_widget(menu_factory, "<main>");
    start_stop_menu_item = gtk_item_factory_get_item(menu_factory, "/Run/Start");
    gui.start_stop_menu_label = GTK_BIN(start_stop_menu_item)->child;
    gui.show_toolbar_menu_item = gtk_item_factory_get_item(menu_factory, "/View/Show Toolbar");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(gui.show_toolbar_menu_item),
                                   dstate.show_toolbar);
    gui.show_sidebar_menu_item = gtk_item_factory_get_item(menu_factory, "/View/Show Sidebar");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(gui.show_sidebar_menu_item),
                                   dstate.show_sidebar);
    gui.show_scrollbars_menu_item = gtk_item_factory_get_item(menu_factory,
                                                              "/View/Show Scrollbars");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(gui.show_scrollbars_menu_item),
                                   dstate.show_scrollbars);
    gui.position_submenu = gtk_item_factory_get_item(menu_factory, "/Position");

    /* Setup recent files list */
    file_menu = gtk_item_factory_get_widget(menu_factory, "/File");
    for (i=0; i < MAX_RECENT_FILES; i++) {
        if (state.recent_files[i].filename)
            str = dsprintf("%d. %s", i+1, state.recent_files[i].filename);
        else
            str = dsprintf("%d.", i+1);
        item = gtk_menu_item_new_with_label(str);
        free(str);
        state.recent_files[i].menu_item = item;
        state.recent_files[i].label     = GTK_BIN(item)->child;
        gtk_label_set_pattern(GTK_LABEL(state.recent_files[i].label), "_");
        gtk_widget_add_accelerator(item, "activate_item",
                                   gtk_menu_get_uline_accel_group(GTK_MENU(file_menu)),
                                   (i+1) + 0x30, 0, GTK_ACCEL_LOCKED);
        gtk_widget_add_accelerator(item, "activate_item", accelerators, (i+1) + 0x30,
                                   GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
        num = safe_malloc(sizeof(int32));
        *num = i;
        gtk_signal_connect(GTK_OBJECT(item), "activate", GTK_SIGNAL_FUNC(file_recent), num);
        gtk_menu_shell_insert(GTK_MENU_SHELL(file_menu), item, RECENT_FILES_INSERT_POS+i);
    }
    /* Insert separator after recent files */
    gui.recent_files_separator = gtk_menu_item_new();
    gtk_menu_shell_insert(GTK_MENU_SHELL(file_menu), gui.recent_files_separator,
                          RECENT_FILES_INSERT_POS+i);

    /* Add menubar to the window */
    gtk_box_pack_start(GTK_BOX(containing_box), gui.menubar, FALSE, FALSE, 0);
}

/* Setup the toolbar, including speed slider and tick display to the right of the buttons
 */
void init_toolbar(GtkWidget* containing_box)
{
    GtkWidget*     toolbar;
    GtkWidget*     icon;
    GtkWidget*     button;
    GtkWidget*     toolbar_frame;
    GtkWidget*     speed_frame;
    GtkWidget*     tick_and_pop_frame;
    GtkWidget*     hbox;
    toolbar_item*  item;

    /* Create the button bar */
    toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
    gtk_toolbar_set_space_size(GTK_TOOLBAR(toolbar), 10);

    /* Add toolbar buttons */
    for (item=toolbar_items; item->description; item++) {
        if (STR_EQUAL(item->description, "sep"))
            gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
        else {
            icon = load_pixmap_from_rgba_array(item->icon, ICON_WIDTH, ICON_HEIGHT,
                                               &(gui.window->style->bg[GTK_STATE_NORMAL]));
            button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), NULL, item->description, NULL,
                                             icon, item->callback, NULL);
            if (item->icon == start_icon) {
                gui.start_stop_toolbar_button = button;
                gui.start_pixmap = icon;
            }
        }
    }
    gui.stop_pixmap = load_pixmap_from_rgba_array(stop_icon, ICON_WIDTH, ICON_HEIGHT,
                                                  &(gui.window->style->bg[GTK_STATE_NORMAL]));

    /* Frame the button bar */
    toolbar_frame = gtk_frame_new(NULL);
    gtk_container_add(GTK_CONTAINER(toolbar_frame), toolbar);

    /* Create and frame the Gen/s slider */
    hbox = gtk_hbox_new(FALSE, 0);
    init_speed_box(hbox);
    speed_frame = gtk_frame_new(NULL);
    gtk_container_add(GTK_CONTAINER(speed_frame), hbox);

    /* Create and frame the tick and population indicators */
    hbox = gtk_hbox_new(FALSE, 0);
    init_tick_display(hbox);
    init_population_display(hbox);
    tick_and_pop_frame = gtk_frame_new(NULL);
    gtk_container_add(GTK_CONTAINER(tick_and_pop_frame), hbox);

    /* Pack the button bar, speed control, and tick display into gui.toolbar_box */
    gui.toolbar_box = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(gui.toolbar_box), toolbar_frame, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(gui.toolbar_box), speed_frame, TRUE, TRUE, 10);
    gtk_box_pack_end(GTK_BOX(gui.toolbar_box), tick_and_pop_frame, FALSE, FALSE, 0);

    /* Add the toolbar box to the container */
    gtk_box_pack_start(GTK_BOX(containing_box), gui.toolbar_box, FALSE, FALSE, 0);
}

/* Setup the speed slider box
 */
void init_speed_box(GtkWidget* containing_box)
{
    GtkWidget*  label;
    GtkObject*  slider_params;
    char*       speed_str;

    /* Setup label */
    label = gtk_label_new(" Speed:  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
    speed_str = dsprintf("%d", state.speed);
    gui.speed_label = gtk_label_new(speed_str);
    free(speed_str);
    set_speed_label_size(config.speed_slider_max);
    gtk_misc_set_alignment(GTK_MISC(gui.speed_label), 1, 0.5);
    gtk_box_pack_start(GTK_BOX(containing_box), gui.speed_label, FALSE, FALSE, 0);
    label = gtk_label_new(" Gen/s  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);

    /* Setup slider */
    slider_params = gtk_adjustment_new(state.speed, config.speed_slider_min,
                                       config.speed_slider_max, 1, config.speed_increment, 0);
    gui.speed_slider = gtk_hscale_new(GTK_ADJUSTMENT(slider_params));
    gtk_scale_set_draw_value(GTK_SCALE(gui.speed_slider), FALSE);
    gtk_signal_connect(GTK_OBJECT(slider_params), "value_changed",
                       GTK_SIGNAL_FUNC(handle_speed_slider_change), NULL);
    gtk_box_pack_start(GTK_BOX(containing_box), gui.speed_slider, TRUE, TRUE, 0);

    /* Extra space at the end */
    label = gtk_label_new("  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
}

/* Set up the tick display
 */
void init_tick_display(GtkWidget* containing_box)
{
    GtkWidget*  label;

    label = gtk_label_new(" Tick:  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
    gui.tick_label = gtk_label_new("0");
    gtk_box_pack_start(GTK_BOX(containing_box), gui.tick_label, FALSE, FALSE, 0);
    label = gtk_label_new("   ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
}

/* Set up the population display
 */
void init_population_display(GtkWidget* containing_box)
{
    GtkWidget*  label;

    label = gtk_label_new("Pop:  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
    gui.population_label = gtk_label_new("0");
    gtk_box_pack_start(GTK_BOX(containing_box), gui.population_label, FALSE, FALSE, 0);
    label = gtk_label_new("  ");
    gtk_box_pack_start(GTK_BOX(containing_box), label, FALSE, FALSE, 0);
}

/* Setup the pattern loader sidebar
 */
void init_sidebar(GtkWidget* containing_box)
{
    char*  header = "Patterns";

    gui.pattern_list = gtk_clist_new_with_titles(1, &header);
    update_pattern_list();
    gtk_clist_column_titles_passive(GTK_CLIST(gui.pattern_list));
    gtk_clist_set_selection_mode(GTK_CLIST(gui.pattern_list), GTK_SELECTION_SINGLE);
    gtk_signal_connect(GTK_OBJECT(gui.pattern_list), "select_row",
                       GTK_SIGNAL_FUNC(handle_sidebar_select), NULL);
    gui.sidebar = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(gui.sidebar), GTK_POLICY_AUTOMATIC,
                                   GTK_POLICY_AUTOMATIC);
    gtk_container_add(GTK_CONTAINER(gui.sidebar), gui.pattern_list);
    gtk_box_pack_start(GTK_BOX(containing_box), gui.sidebar, FALSE, FALSE, 0);
}

/* Setup the Life pattern drawing area
 */
void init_canvas(GtkWidget* containing_box)
{
    GtkWidget*  vbox;
    GtkWidget*  hbox;
    GtkObject*  scroll_params;

    vbox = gtk_vbox_new(FALSE, 0);
    hbox = gtk_hbox_new(FALSE, 0);

    /* Create and frame the canvas */
    gui.canvas = gtk_drawing_area_new();
    gtk_signal_connect(GTK_OBJECT(gui.canvas), "configure_event",
                       GTK_SIGNAL_FUNC(handle_canvas_resize), NULL);
    gtk_signal_connect(GTK_OBJECT(gui.canvas), "expose_event",
                       GTK_SIGNAL_FUNC(handle_canvas_expose), NULL);
    gtk_signal_connect(GTK_OBJECT(gui.canvas), "button_press_event",
                       GTK_SIGNAL_FUNC(handle_mouse_press), NULL);
    gtk_signal_connect(GTK_OBJECT(gui.canvas), "button_release_event",
                       GTK_SIGNAL_FUNC(handle_mouse_release), NULL);
    gtk_widget_set_events(gui.canvas, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
    gtk_box_pack_start(GTK_BOX(hbox), gui.canvas, TRUE, TRUE, 0);

    /* Create a vertical scrollbar for the canvas. A number of these settings won't get properly
     * set until handle_canvas_resize is called, so use dummy values for now.
     */
    scroll_params = gtk_adjustment_new(INIT_Y_CENTER, 0, WORLD_SIZE-1, 1, 1, 0);
    gui.vscrollbar = gtk_vscrollbar_new(GTK_ADJUSTMENT(scroll_params));
    gtk_box_pack_start(GTK_BOX(hbox), gui.vscrollbar, FALSE, FALSE, 0);

    /* add the hbox to the containing box */
    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

    /* Create and add a horizontal scrollbar for the canvas. */
    scroll_params = gtk_adjustment_new(INIT_X_CENTER, 0, WORLD_SIZE-1, 1, 1, 0);
    gui.hscrollbar = gtk_hscrollbar_new(GTK_ADJUSTMENT(scroll_params));
    gtk_box_pack_start(GTK_BOX(vbox), gui.hscrollbar, FALSE, FALSE, 0);

    /* Put the canvas+scrollbars into the box */
    gtk_box_pack_start(GTK_BOX(containing_box), vbox, TRUE, TRUE, 0);
}

/*** Misc. GUI functions ***/

/* Adjust horizontal and vertical scrollbars for the current viewport start and size.
 */
void adjust_scrollbars(void)
{
    GtkObject*  scroll_params;

    /* Necessary to say WORLD_SIZE instead of WORLD_SIZE-1 to get full range. Why? */

    scroll_params = gtk_adjustment_new(dstate.viewport.start.x, 0, WORLD_SIZE,
                                       dstate.viewport_size.width/8, dstate.viewport_size.width,
                                       dstate.viewport_size.width);
    gtk_range_set_adjustment(GTK_RANGE(gui.hscrollbar), GTK_ADJUSTMENT(scroll_params));
    gtk_signal_connect(GTK_OBJECT(scroll_params), "value_changed",
                       GTK_SIGNAL_FUNC(handle_hscrollbar_change), NULL);

    scroll_params = gtk_adjustment_new(dstate.viewport.start.y, 0, WORLD_SIZE,
                                       dstate.viewport_size.height/8, dstate.viewport_size.height,
                                       dstate.viewport_size.height);
    gtk_range_set_adjustment(GTK_RANGE(gui.vscrollbar), GTK_ADJUSTMENT(scroll_params));
    gtk_signal_connect(GTK_OBJECT(scroll_params), "value_changed",
                       GTK_SIGNAL_FUNC(handle_vscrollbar_change), NULL);
    gtk_widget_queue_draw(gui.hscrollbar);
    gtk_widget_queue_draw(gui.vscrollbar);
}

/* Adjust horizontal and vertical scrollbars for the current viewport start */
void adjust_scrollbar_values(void)
{
    GtkAdjustment*  sb_state;

    sb_state = gtk_range_get_adjustment(GTK_RANGE(gui.hscrollbar));
    gtk_adjustment_set_value(sb_state, dstate.viewport.start.x);
    sb_state = gtk_range_get_adjustment(GTK_RANGE(gui.vscrollbar));
    gtk_adjustment_set_value(sb_state, dstate.viewport.start.y);
}

/* Display an error/warning dialog with the given message and an "Okay" button. The message may be
 * specified in printf-style fashion, with format string and arguments.
 */
void error_dialog(const char* format, ...)
{
    va_list     args;
    GtkWidget*  dialog;
    GtkWidget*  message;
    GtkWidget*  hbox;
    GtkWidget*  error_pixmap;
    GtkWidget*  ok_button;
    char*       str;

    va_start(args, format);
    str = vdsprintf(format, args);
    va_end(args);

    /* Set up the dialog window */
    dialog = gtk_dialog_new();
    gtk_window_set_title(GTK_WINDOW(dialog), TITLE " Error");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(dialog), 5);
    gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 3);

    hbox = gtk_hbox_new(FALSE, 0);

    /* Set up the message with an error icon next to it */
    error_pixmap = load_pixmap_from_rgba_array(error_icon, ICON_WIDTH, ICON_HEIGHT,
                                               &(gui.window->style->bg[GTK_STATE_NORMAL]));
    gtk_box_pack_start(GTK_BOX(hbox), error_pixmap, FALSE, FALSE, 5);
    message = gtk_label_new(str);
    gtk_label_set_justify(GTK_LABEL(message), GTK_JUSTIFY_LEFT);
    gtk_box_pack_start(GTK_BOX(hbox), message, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);

    /* Set up the "okay" button */
    ok_button = gtk_button_new_with_label(" Okay ");
    gtk_signal_connect_object(GTK_OBJECT(ok_button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), ok_button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(ok_button);

    free(str);
    /* Show the dialog */
    gtk_widget_show_all(dialog);
}

/* Redraw the entire Life pixmap, then trigger an expose event on the canvas. Also set
 * state.skipped_frames back to 0.
 */
void full_canvas_redraw(void)
{
    state.skipped_frames = 0;
    draw_life_pixmap();
    gtk_widget_queue_draw(gui.canvas);
}

/* Handle a mouse button click on the Life canvas.
 */
void handle_mouse_press(GtkWidget *widget, GdkEventButton *event)
{
    point  pt, lpt;
    rect   r;

    pt.x = event->x;
    pt.y = event->y;
    get_logical_coords(&pt, &lpt);
    if (state.tracking_mouse == PASTING && event->button < 3)
        /* They've chosen where to paste a copied region */
        edit_paste_win_style(&lpt);
    else if (event->button == 1) {
        /* If they held the shift key down when clicking, start selecting a region. Otherwise
         * start drawing. */
        if (event->state & GDK_SHIFT_MASK) {
            state.tracking_mouse = SELECTING;
            state.selection_start_time = get_time_milliseconds();
            r.start = r.end = lpt;
            screen_box_update(&state.selection, &r);
        } else
            user_draw(&lpt);
    } else if (event->button == 2)
        /* They're pasting a region unix-style by clicking the middle mouse button */
        edit_paste_unix_style(&lpt);
    else if (event->button == 3) {
        /* Recenter at this point */
        set_viewport_position(&lpt, TRUE);
        adjust_scrollbar_values();
        full_canvas_redraw();
    } else if (event->button == 4)
        position_scroll_up();
    else if (event->button == 5)
        position_scroll_down();
}

/* Handle a mouse button release on the Life canvas.
 */
void handle_mouse_release(GtkWidget *widget, GdkEventButton *event)
{
    if (event->button == 1 &&
        (state.tracking_mouse == DRAWING || state.tracking_mouse == SELECTING)) {
        /* If they just quickly shift-clicked the left mouse button, deactivate the current
         * selection. */
        if (state.tracking_mouse == SELECTING &&
            state.selection.start.x == state.selection.end.x &&
            state.selection.start.y == state.selection.end.y &&
            get_time_milliseconds() - state.selection_start_time < 250) {
            screen_box_update(&state.selection, &null_rect);
        }
        state.tracking_mouse = NOT_TRACKING;
        state.last_drawn = null_point;
    }
}

/* Handle mouse movement over the Life canvas (dragging to draw, dragging to select, or choosing a
 * paste region). This is not actually an event handler, but is called manually from the main loop
 * if state.tracking_mouse is true.
 */
void handle_mouse_movement(void)
{
    GdkModifierType  mouse_mask;
    point  pt, lpt;
    rect   r;
    int32  scroll_increment;

    gdk_window_get_pointer(gui.canvas->window, &pt.x, &pt.y, &mouse_mask);
    if (state.tracking_mouse != PASTING && !(mouse_mask & GDK_BUTTON1_MASK)) {
        /* They released the mouse button while we weren't looking */
        state.tracking_mouse = NOT_TRACKING;
        state.last_drawn = null_point;
        return;
    }

    /* If they're at or off the end of the canvas, scroll with them */
    scroll_increment = MAX_ZOOM / dstate.zoom;
    if (pt.x <= 0)
        position_scroll_generic(-1, 0, scroll_increment);
    else if (pt.x >= dstate.eff_canvas_size.width-1)
        position_scroll_generic(1, 0, scroll_increment);
    if (pt.y <= 0)
        position_scroll_generic(0, -1, scroll_increment);
    else if (pt.y >= dstate.eff_canvas_size.height-1)
        position_scroll_generic(0, 1, scroll_increment);

    get_logical_coords(&pt, &lpt);

    if (state.tracking_mouse == DRAWING) {
        if (points_identical(&state.last_drawn, &null_point))
            user_draw(&lpt);
        else if (!points_identical(&lpt, &state.last_drawn))
            draw_from_to(&state.last_drawn, &lpt);
    } else if (state.tracking_mouse == SELECTING) {
        r.start = state.selection.start;
        r.end   = lpt;
        screen_box_update(&state.selection, &r);
    } else {  /* pasting */
        r.start = lpt;
        r.end.x = r.start.x + (state.copy_rect.end.x - state.copy_rect.start.x);
        r.end.y = r.start.y + (state.copy_rect.end.y - state.copy_rect.start.y);
        if (r.end.x >= WORLD_SIZE || r.end.y >= WORLD_SIZE)
            r = null_rect;
        screen_box_update(&state.paste_box, &r);
    }
}

/* Called when all or some of the Life display needs to be drawn onscreen. Redraw the exposed
 * rectangle from the offscreen pixmap.
 */
void handle_canvas_expose(GtkWidget *widget, GdkEventExpose *event)
{
    gdk_draw_indexed_image(gui.canvas->window,
                     gui.canvas->style->fg_gc[GTK_WIDGET_STATE(gui.canvas)],
                     event->area.x, event->area.y, event->area.width, event->area.height,
                     GDK_RGB_DITHER_NONE,
                     dstate.life_pixmap + event->area.y*dstate.canvas_size.width + event->area.x,
                     dstate.canvas_size.width, dstate.life_pixmap_colormap);
}

/* Called when the window dimensions are changed (including at startup): Set up various globals
 * dependent on window size, and recreate and redraw the canvas pixmap. Since an expose event will
 * already have been generated, we don't generate one here.
 */
void handle_canvas_resize(void)
{
    point  center;
    int32  i;

    /* Get canvas dimensions (in pixels) */
    dstate.canvas_size.width  = gui.canvas->allocation.width;
    dstate.canvas_size.height = gui.canvas->allocation.height;

    /* Determine viewport dimensions (in cells) at all zoom levels for the new window size */
    for (i=MIN_ZOOM; i <= MAX_ZOOM; i *= 2) {
        if (DRAW_GRID(i)) {
            dstate.viewport_sizes[i].width  = (dstate.canvas_size.width  - 1) / (i + 1);
            dstate.viewport_sizes[i].height = (dstate.canvas_size.height - 1) / (i + 1);
        }
        else {
            dstate.viewport_sizes[i].width  = dstate.canvas_size.width  / i;
            dstate.viewport_sizes[i].height = dstate.canvas_size.height / i;
        }
    }

    /* Determine the original viewport center point. If the program has just started up, pick the
     * middle of the world.
     */
    if (!dstate.life_pixmap) {   /* program just started */
        center.x = INIT_X_CENTER;
        center.y = INIT_Y_CENTER;
    } else {
        center.x = dstate.viewport.start.x + dstate.viewport_size.width/2;
        center.y = dstate.viewport.start.y + dstate.viewport_size.height/2;
    }

    /* Record new viewport dimensions at the current zoom level */
    dstate.viewport_size.width  = dstate.viewport_sizes[dstate.zoom].width;
    dstate.viewport_size.height = dstate.viewport_sizes[dstate.zoom].height;

    /* Determine the new effective canvas size */
    set_effective_canvas_size();

    /* Set new viewport start to center around the same point as before */
    set_viewport_position(&center, TRUE);

    /* Adjust scrollbars for the new viewport size and position */
    adjust_scrollbars();

    /* Create and draw a new pixmap for the new viewport */
    if (dstate.canvas_size.width * dstate.canvas_size.height > dstate.life_pixmap_alloc_len) {
        dstate.life_pixmap_alloc_len = dstate.canvas_size.width * dstate.canvas_size.height;
        dstate.life_pixmap = safe_realloc(dstate.life_pixmap, dstate.life_pixmap_alloc_len);
    }
    draw_life_pixmap();
}

/* Called when the horizontal scrollbar is adjusted (by the user or by our program). Adjust the
 * viewport start, and do a full redraw.
 */
void handle_hscrollbar_change(void)
{
    GtkAdjustment*  sb_state;
    point           pt;

    sb_state = gtk_range_get_adjustment(GTK_RANGE(gui.hscrollbar));
    pt.x = ROUND(sb_state->value);
    pt.y = dstate.viewport.start.y;
    set_viewport_position(&pt, FALSE);
    full_canvas_redraw();
}

/* Called when the vertical scrollbar is adjusted (by the user or by our program). Adjust the
 * viewport start, and do a full redraw.
 */
void handle_vscrollbar_change(void)
{
    GtkAdjustment*  sb_state;
    point           pt;

    sb_state = gtk_range_get_adjustment(GTK_RANGE(gui.vscrollbar));
    pt.x = dstate.viewport.start.x;
    pt.y = ROUND(sb_state->value);
    set_viewport_position(&pt, FALSE);
    full_canvas_redraw();
}

/* Called when the user selects a pattern file from the sidebar */
void handle_sidebar_select(GtkCList *clist, gint row, gint column, GdkEventButton *event,
                           gpointer user_data)
{
    char*  filename;
    char*  path;

    if (row < 0)
        return;
    if (gtk_clist_get_text(GTK_CLIST(clist), row, column, &filename)) {
        path = dsprintf("%s%s.lif", config.patterns_dir, filename);
        attempt_load_pattern(path);
        free(path);
    }
}

/* Called when the speed slider has been adjusted (by the user or by our program). Set the new
 * speed, update the speed label, and reset the FPS measurement if the pattern is running.
 */
void handle_speed_slider_change(void)
{
    GtkAdjustment*  slider_state;
    char*           speed_str;

    slider_state = gtk_range_get_adjustment(GTK_RANGE(gui.speed_slider));
    state.speed = ROUND(slider_state->value);
    speed_str = dsprintf("%d", state.speed);
    gtk_label_set_text(GTK_LABEL(gui.speed_label), speed_str);
    free(speed_str);

    /* Reset the running FPS measurement if necessary */
    if (state.pattern_running) {
        state.start_tick = tick;
        state.start_time = get_time_milliseconds();
    }
}

/* Called when the zoom level has changed by some means. Update various globals dependent on zoom
 * level, then do a full redraw.
 */
void handle_zoom_change(void)
{
    point  center;

    /* Determine the original viewport center point */
    center.x = dstate.viewport.start.x + dstate.viewport_size.width/2;
    center.y = dstate.viewport.start.y + dstate.viewport_size.height/2;

    /* Record new viewport dimensions at the current zoom level */
    dstate.viewport_size.width  = dstate.viewport_sizes[dstate.zoom].width;
    dstate.viewport_size.height = dstate.viewport_sizes[dstate.zoom].height;

    /* Determine the effective canvas size for the new zoom level */
    set_effective_canvas_size();

    /* Set new viewport start to center around the same point as before */
    set_viewport_position(&center, TRUE);

    /* Adjust scrollbars for the new viewport size and position */
    adjust_scrollbars();

    /* Draw the pixmap and generate a canvas expose */
    full_canvas_redraw();
}

/* Put the current filename (taken from state.recent_files[0]) into the main window title */
void put_filename_in_window_title(void)
{
    char*  title;

    title = dsprintf("%s: %s", TITLE, state.recent_files[0].filename);
    gtk_window_set_title(GTK_WINDOW(gui.window), title);
    free(title);
}

/* Record the current effective canvas size (canvas size minus any unused space) */
void set_effective_canvas_size(void)
{
    if (DRAW_GRID(dstate.zoom)) {
        dstate.eff_canvas_size.width  = dstate.viewport_size.width  * (dstate.zoom+1) + 1;
        dstate.eff_canvas_size.height = dstate.viewport_size.height * (dstate.zoom+1) + 1;
    } else {
        dstate.eff_canvas_size.width  = dstate.viewport_size.width  * dstate.zoom;
        dstate.eff_canvas_size.height = dstate.viewport_size.height * dstate.zoom;
    }
}

/* Adjust horizontal widget size for the speed label, based on max speed */
void set_speed_label_size(int32 max_speed)
{
    char*  speed_str;

    speed_str = dsprintf("%d", max_speed);
    gtk_widget_set_usize(GTK_WIDGET(gui.speed_label),
                         gdk_string_measure(gui.window->style->font, speed_str) + 2, 0);
    free(speed_str);
}

/* Set the global variables for viewport location, and let the backend know about the new
 * viewport. If center_around is true, attempt to center around the given point, otherwise try to
 * set the given point at the upper left corner. If the viewport would go off the edge of the
 * world, readjust it.
 */
void set_viewport_position(const point* pt, boolean center_around)
{
    /* Set viewport start */
    dstate.viewport.start.x = pt->x;
    dstate.viewport.start.y = pt->y;
    if (center_around) {
        dstate.viewport.start.x -= dstate.viewport_size.width/2;
        dstate.viewport.start.y -= dstate.viewport_size.height/2;
    }

    /* Readjust viewport start if the current viewport goes off the edge of the world */
    if (dstate.viewport.start.x < 0)
        dstate.viewport.start.x = 0;
    else if (dstate.viewport.start.x + dstate.viewport_size.width > WORLD_SIZE)
        dstate.viewport.start.x = WORLD_SIZE - dstate.viewport_size.width;
    if (dstate.viewport.start.y < 0)
        dstate.viewport.start.y = 0;
    else if (dstate.viewport.start.y + dstate.viewport_size.height > WORLD_SIZE)
        dstate.viewport.start.y = WORLD_SIZE - dstate.viewport_size.height;

    /* Set viewport end */
    dstate.viewport.end.x = dstate.viewport.start.x + dstate.viewport_size.width  - 1;
    dstate.viewport.end.y = dstate.viewport.start.y + dstate.viewport_size.height - 1;

    /* Let the backend know about the new viewport */
    set_viewport(dstate.viewport.start.x, dstate.viewport.start.y, dstate.viewport.end.x,
                 dstate.viewport.end.y);
}

/* Trigger an expose event on the portion of the Life canvas which has changed over the last tick,
 * setting state.skipped_frames back to 0.
 */
void trigger_canvas_update(void)
{
    dimension  size;

    state.skipped_frames = 0;
    get_rect_dimensions(&dstate.update, &size);
    if (size.width > 0 && size.height > 0)
        gtk_widget_queue_draw_area(gui.canvas, dstate.update.start.x, dstate.update.start.y,
                                   size.width, size.height);
}

/* Update the onscreen population label to match the current population. */
void update_population_label(void)
{
    char  str[20];

    sprintf(str, "%u", population);
    gtk_label_set_text(GTK_LABEL(gui.population_label), str);
}

/* Regenerate the list of patterns for the sidebar. */
void update_pattern_list(void)
{
    DIR*            dir;
    struct dirent*  file;
    char*           filename;
    int32           num_patterns = 0;
    int32           ext_len;
    int32           len;

    gtk_clist_freeze(GTK_CLIST(gui.pattern_list));
    gtk_clist_clear(GTK_CLIST(gui.pattern_list));

    ext_len = strlen(DEFAULT_EXTENSION);
    dir = opendir(config.patterns_dir);
    if (dir) {
        while ((file = readdir(dir)) != NULL) {
            filename = safe_strdup(file->d_name);
            len = strlen(filename);
            if (len > ext_len + 1 && STR_EQUAL(filename + len - ext_len, DEFAULT_EXTENSION)) {
                filename[len - ext_len - 1] = '\0';
                gtk_clist_append(GTK_CLIST(gui.pattern_list), &filename);
                num_patterns++;
            }
            free(filename);
        }
        closedir(dir);
    }

    gtk_clist_sort(GTK_CLIST(gui.pattern_list));
    gtk_widget_set_usize(gui.pattern_list,
                         gtk_clist_optimal_column_width(GTK_CLIST(gui.pattern_list), 0) + 40, 0);

    gtk_clist_thaw(GTK_CLIST(gui.pattern_list));
}

/* Update the onscreen tick label to match the current tick. */
void update_tick_label(void)
{
    char  str[20];

    sprintf(str, "%u", tick);
    gtk_label_set_text(GTK_LABEL(gui.tick_label), str);
}

/* Create an integral, snap-to-ticks spinbutton with the given initial, minimum and maximum values.
 */
GtkWidget* create_spinbox(int32 init_value, int32 min, int32 max)
{
    GtkWidget*  spinbox;
    GtkObject*  adjustment;
    char*       max_str;

    adjustment = gtk_adjustment_new(init_value, min, max, 1, 1, 0);
    spinbox = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 1, 0);
    gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(spinbox), TRUE);
    gtk_spin_button_set_snap_to_ticks(GTK_SPIN_BUTTON(spinbox), TRUE);
    max_str = dsprintf("%d", max);
    gtk_widget_set_usize(spinbox, gdk_string_measure(gui.window->style->font, max_str) + 25, 0);
    free(max_str);

    return spinbox;
}

/*** Functions for drawing the offscreen pixmap ***/

/* Fully redraw the Life pixmap: backdrop, grid, selection box, and live cells.
 */
void draw_life_pixmap(void)
{
    cage_type*  c;
    int32       x, y;

    /* Draw the backdrop */
    memset(dstate.life_pixmap, BG_COLOR_INDEX, dstate.canvas_size.width*dstate.canvas_size.height);

    /* Draw the live cells */
    while ((c = loop_cages_onscreen()) != NULL) {
        x = c->x * CAGE_SIZE;
        y = c->y * CAGE_SIZE;
        if (c->bnw[parity])
            draw_life_block(c->bnw[parity], x, y, TRUE);
        if (c->bne[parity])
            draw_life_block(c->bne[parity], x + BLOCK_SIZE, y, TRUE);
        if (c->bsw[parity])
            draw_life_block(c->bsw[parity], x, y + BLOCK_SIZE, TRUE);
        if (c->bse[parity])
            draw_life_block(c->bse[parity], x + BLOCK_SIZE, y + BLOCK_SIZE, TRUE);
    }

    draw_grid_and_boxes();
}

/* Redraw just the zoom grid and boxes (selection, paste) on the life pixmap, if applicable.
 */
void draw_grid_and_boxes(void)
{
    int32  x, y;

    /* Draw the grid, if necessary */
    if (DRAW_GRID(dstate.zoom)) {
        /* Draw the horizontal lines */
        for (y=0; y < dstate.canvas_size.height; y += (dstate.zoom+1))
            memset(dstate.life_pixmap + y*dstate.canvas_size.width, GRID_COLOR_INDEX,
                   dstate.eff_canvas_size.width);
        /* Draw the vertical lines */
        for (x=0; x < dstate.canvas_size.width; x += (dstate.zoom+1)) {
            for (y=0; y < dstate.eff_canvas_size.height; y++)
                dstate.life_pixmap[y*dstate.canvas_size.width + x] = GRID_COLOR_INDEX;
        }
    }

    /* Draw the selection box and/or paste-indicator box */
    draw_screen_box(&state.selection);
    draw_screen_box(&state.paste_box);
}

/* Draw the given box (selection or paste indicator) to the pixmap.
 */
void draw_screen_box(const rect* box)
{
    dimension  boxsize;
    rect       nbox;
    rect       r, cr;
    int32      y;

    if (box->start.x < 0)
        return;
    nbox = *box;
    normalize_rectangle(&nbox);
    if (nbox.start.x > dstate.viewport.end.x || nbox.start.y > dstate.viewport.end.y ||
        nbox.end.x < dstate.viewport.start.x || nbox.end.y < dstate.viewport.start.y) {
        return;
    }

    get_screen_rectangle(&nbox, &r);
    r.end.x += dstate.zoom-1;
    r.end.y += dstate.zoom-1;
    if (DRAW_GRID(dstate.zoom)) {
        r.start.x--;
        r.start.y--;
        r.end.x++;
        r.end.y++;
    }

    cr.start.x = MAX(0, r.start.x);
    cr.start.y = MAX(0, r.start.y);
    cr.end.x   = MIN(dstate.eff_canvas_size.width-1,  r.end.x);
    cr.end.y   = MIN(dstate.eff_canvas_size.height-1, r.end.y);
    get_rect_dimensions(&cr, &boxsize);

    if (r.start.y >= 0)
        memset(dstate.life_pixmap + r.start.y*dstate.canvas_size.width + cr.start.x,
               SELECT_COLOR_INDEX, boxsize.width);
    if (r.end.y < dstate.eff_canvas_size.height)
        memset(dstate.life_pixmap + r.end.y*dstate.canvas_size.width + cr.start.x,
               SELECT_COLOR_INDEX, boxsize.width);
    if (r.start.x >= 0) {
        for (y=cr.start.y; y <= cr.end.y; y++)
            dstate.life_pixmap[y*dstate.canvas_size.width + r.start.x] = SELECT_COLOR_INDEX;
    }
    if (r.end.x < dstate.eff_canvas_size.width) {
        for (y=cr.start.y; y <= cr.end.y; y++)
            dstate.life_pixmap[y*dstate.canvas_size.width + r.end.x] = SELECT_COLOR_INDEX;
    }
}

/* Draw the given 4x4 cell block onto the Life pixmap, at the given starting coordinates.
 * If full_update is TRUE, draw only live cells (because the screen has been cleared beforehand),
 * otherwise draw all cells. If full_update is FALSE, update the current clipping rectangle
 * (dstate.update.start.x, etc.) to include this block. This function may be called by the
 * backend.
 */
void draw_life_block(uint16 block, int32 xstart, int32 ystart, boolean full_update)
{
    int32     xend, yend;
    int32     minx, miny, maxx, maxy;
    int32     real_minx, real_miny, real_maxx, real_maxy;
    int32     realx, realy;
    int32     grid_block_size;
    int32     start_bit;
    int32     bit;
    int32     x, y, i;
    uint8     color;
    uint8*    spos;
    uint8*    pos;

    grid_block_size = (DRAW_GRID(dstate.zoom) ? dstate.zoom+1 : dstate.zoom);

    /* Figure out the screen offset, in blocks */
    xstart -= dstate.viewport.start.x;
    ystart -= dstate.viewport.start.y;
    xend = xstart + BLOCK_SIZE - 1;
    yend = ystart + BLOCK_SIZE - 1;
    minx = MAX(xstart, 0);
    miny = MAX(ystart, 0);
    maxx = MIN(xend, dstate.viewport_size.width-1);
    maxy = MIN(yend, dstate.viewport_size.height-1);
    if (minx > maxx || miny > maxy)
        return;

    /* Update the clipping rectangle, if necessary */
    if (!full_update) {
        real_minx = minx * grid_block_size;
        real_miny = miny * grid_block_size;
        real_maxx = maxx * grid_block_size + (dstate.zoom - 1);
        real_maxy = maxy * grid_block_size + (dstate.zoom - 1);
        if (DRAW_GRID(dstate.zoom)) {
            real_minx++;
            real_miny++;
            real_maxx++;
            real_maxy++;
        }
        if (real_minx < dstate.update.start.x)
            dstate.update.start.x = real_minx;
        if (real_miny < dstate.update.start.y)
            dstate.update.start.y = real_miny;
        if (real_maxx > dstate.update.end.x)
            dstate.update.end.x = real_maxx;
        if (real_maxy > dstate.update.end.y)
            dstate.update.end.y = real_maxy;
    }

    start_bit = (miny - ystart) * BLOCK_SIZE + (minx - xstart);
    /* Draw the cells */
    if (dstate.zoom > 1) {
        for (y=miny; y <= maxy; y++, start_bit += BLOCK_SIZE) {
            for (x=minx, bit=start_bit; x <= maxx; x++, bit++) {
                color = ((block & (1 << bit)) ? CELL_COLOR_INDEX : BG_COLOR_INDEX);
                if (full_update && color == BG_COLOR_INDEX)
                    continue;
                realx = x * grid_block_size;
                realy = y * grid_block_size;
                if (DRAW_GRID(dstate.zoom)) {
                    realx++;
                    realy++;
                }
                for (i=0, pos=dstate.life_pixmap + realy*dstate.canvas_size.width + realx;
                     i < dstate.zoom;
                     i++, pos += dstate.canvas_size.width)
                    memset(pos, color, dstate.zoom);
            }
        }
    } else {
        spos = dstate.life_pixmap + miny * dstate.canvas_size.width + minx;
        for (y=miny; y <= maxy; y++, spos += dstate.canvas_size.width, start_bit += BLOCK_SIZE) {
            for (x=minx, bit=start_bit, pos=spos; x <= maxx; x++, pos++, bit++) {
                color = ((block & (1 << bit)) ? CELL_COLOR_INDEX : BG_COLOR_INDEX);
                *pos = color;
            }
        }
    }
}

/* Set/unset the cell at the given absolute (logical) coordinates, then update the Life state and
 * the screen. If the coordinate are outside of the viewport, adjust them. If the user just clicked
 * the mouse, check the color of the cell to determine whether to set or erase. If they've been
 * dragging the mouse, continue either setting or erasing.
 */
void user_draw(const point* pt)
{
    static uint8  color;
    point   npt;
    point   spt;
    uint8*  pos;
    uint8   cur_color;
    int32   i;

    npt = *pt;
    if (npt.x < dstate.viewport.start.x)
        npt.x = dstate.viewport.start.x;
    else if (npt.x > dstate.viewport.end.x)
        npt.x = dstate.viewport.end.x;
    if (npt.y < dstate.viewport.start.y)
        npt.y = dstate.viewport.start.y;
    else if (npt.y > dstate.viewport.end.y)
        npt.y = dstate.viewport.end.y;
    state.last_drawn = npt;
    get_screen_coords(&npt, &spt);

    cur_color = dstate.life_pixmap[spt.y*dstate.canvas_size.width + spt.x];
    if (state.tracking_mouse == DRAWING) {
        if (cur_color == color)
            return;
    } else
        color = ((cur_color == CELL_COLOR_INDEX) ? BG_COLOR_INDEX : CELL_COLOR_INDEX);

    draw_cell(npt.x, npt.y, (color == BG_COLOR_INDEX ? DRAW_UNSET : DRAW_SET));
    if (dstate.zoom == 1)
        dstate.life_pixmap[spt.y*dstate.canvas_size.width + spt.x] = color;
    else {
        for (i=0, pos = dstate.life_pixmap + spt.y*dstate.canvas_size.width + spt.x;
             i < dstate.zoom;
             i++, pos += dstate.canvas_size.width)
            memset(pos, color, dstate.zoom);
    }

    gtk_widget_queue_draw_area(gui.canvas, spt.x, spt.y, dstate.zoom, dstate.zoom);
    state.tracking_mouse = DRAWING;

    update_population_label();
}

/* Call user_draw repeatedly to draw/erase a line between the points start and end.
 */
void draw_from_to(const point* start, const point* end)
{
    point   s, e, pt;
    point   temp;
    int32   xlen, ylen;
    int32   steps, step_len, steps_taken;
    int32   i;

    s = *start;
    e = *end;
    xlen = abs(e.x - s.x) + 1;
    ylen = abs(e.y - s.y) + 1;

    if (xlen > ylen) {
        steps = ylen;
        step_len = xlen / steps;
        if (s.x > e.x) {
            temp = s;
            s = e;
            e = temp;
        }
        steps_taken = 1;
        for (pt=s, i=0; pt.x <= e.x; pt.x++, i++) {
            if (i == step_len && steps_taken < steps) {
                i = 0;
                steps_taken++;
                (s.y < e.y) ? pt.y++ : pt.y--;
            }
            user_draw(&pt);
        }
    } else {
        steps = xlen;
        step_len = ylen / steps;
        if (s.y > e.y) {
            temp = s;
            s = e;
            e = temp;
        }
        steps_taken = 1;
        for (pt=s, i=0; pt.y <= e.y; pt.y++, i++) {
            if (i == step_len && steps_taken < steps) {
                i = 0;
                steps_taken++;
                (s.x < e.x) ? pt.x++ : pt.x--;
            }
            user_draw(&pt);
        }
    }

    state.last_drawn = *end;
}

/*** Bound Functions ***/

void file_new(void)
{
    if (state.pattern_running)
        start_stop();
    state.tracking_mouse = NOT_TRACKING;
    state.selection = state.paste_box = null_rect;
    state.last_drawn = null_point;
    state.moving = FALSE;
    clear_world();
    view_recenter();   /* will trigger a canvas redraw */
    update_tick_label();
    update_population_label();
    sidebar_unselect();
    free(state.pattern_path);
    state.pattern_path = NULL;
    gtk_window_set_title(GTK_WINDOW(gui.window), TITLE);
}

void file_open(void)
{
    GtkWidget*  file_selector;

    file_selector = gtk_file_selection_new("Load Pattern");
    gtk_window_set_position(GTK_WINDOW(file_selector), GTK_WIN_POS_CENTER);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selector), state.current_dir);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->ok_button), "clicked",
                       GTK_SIGNAL_FUNC(file_open_ok), file_selector);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->cancel_button),
                              "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(file_selector));
    gtk_widget_show(file_selector);
}

void file_reopen(void)
{
    if (state.pattern_path)
        attempt_load_pattern(state.pattern_path);
    else
        error_dialog("Reopen failed: no current file");
}

void file_save(void)
{
    if (state.pattern_path)
        attempt_save_pattern(state.pattern_path);
    else
        file_save_as();
}

void file_save_as(void)
{
    GtkWidget*  file_selector;
    char*       home_dir;

    file_selector = gtk_file_selection_new("Save Pattern");
    gtk_window_set_position(GTK_WINDOW(file_selector), GTK_WIN_POS_CENTER);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->ok_button), "clicked",
                       GTK_SIGNAL_FUNC(file_save_as_ok), file_selector);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->cancel_button),
                              "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(file_selector));
    if (state.pattern_path)
        gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selector), state.pattern_path);
    else {
        home_dir = getenv("HOME");
        if (home_dir) {
            home_dir = dsprintf("%s/", home_dir);
            gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selector), home_dir);
            free(home_dir);
        }
    }
    gtk_widget_show(file_selector);
}

void file_description(void)
{
    GtkWidget*  dialog;
    GtkWidget*  textbox;
    GtkWidget*  scrollbar;
    GtkWidget*  button;
    GtkWidget*  alignment;
    GtkWidget*  hbox;
    desc_dialog_info*  info;
    int32  desc_width = 0;
    int32  desc_height;
    int32  w;
    int32  i;

    info = safe_malloc(sizeof(desc_dialog_info));

    /* Determine the width and height of the current description text */
    for (i=0; i < desc_num_lines; i++) {
        w = gdk_string_measure(gui.window->style->font, pattern_description[i]) + 40;
        if (w > desc_width)
            desc_width = w;
    }
    if (desc_width == 0) {
        desc_width = 500;
        desc_height = 200;
    } else {
        if (desc_width < 300)
            desc_width = 300;
        desc_height = MIN(400, 20 * desc_num_lines);
    }

    /* Set up the dialog window */
    dialog = gtk_dialog_new();
    gtk_window_set_title(GTK_WINDOW(dialog), "View/Edit Pattern Description");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_window_set_default_size(GTK_WINDOW(dialog), desc_width, desc_height);
    gtk_container_set_border_width(GTK_CONTAINER(dialog), 5);
    gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 3);
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", GTK_SIGNAL_FUNC(file_description_destroy),
                       info);

    hbox = gtk_hbox_new(FALSE, 0);

    /* Set up the textbox with vertical scrollbar */
    textbox = gtk_text_new(NULL, NULL);
    gtk_text_set_editable(GTK_TEXT(textbox), TRUE);
    gtk_text_set_word_wrap(GTK_TEXT(textbox), TRUE);
    for (i=0; i < desc_num_lines; i++) {
        gtk_text_insert(GTK_TEXT(textbox), NULL, NULL, NULL, pattern_description[i], -1);
        gtk_text_insert(GTK_TEXT(textbox), NULL, NULL, NULL, "\n", -1);
    }
    gtk_box_pack_start(GTK_BOX(hbox), textbox, TRUE, TRUE, 0);
    scrollbar = gtk_vscrollbar_new(GTK_TEXT(textbox)->vadj);
    gtk_box_pack_start(GTK_BOX(hbox), scrollbar, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);

    /* Data to be sent to signal handlers for this dialog */
    info->dialog  = dialog;
    info->textbox = textbox;

    /* Set up the buttons */
    button = gtk_button_new_with_label(" Okay ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(file_description_ok), info);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(button);
    button = gtk_button_new_with_label(" Apply ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(file_description_apply),
                       info);
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);
    button = gtk_button_new_with_label(" Cancel ");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);

    /* Show the dialog */
    gtk_widget_show_all(dialog);
}

void file_recent(GtkWidget* widget, int32* num)
{
    if (state.recent_files[*num].full_path) {
        if (attempt_load_pattern(state.recent_files[*num].full_path))
            sidebar_unselect();
    }
}

void file_quit(void)
{
    save_keybindings();
    save_recent_files_list();
    exit(0);
}

void view_recenter(void)
{
    point  pt;

    pt.x = pt.y = WORLD_SIZE/2;
    set_viewport_position(&pt, TRUE);
    adjust_scrollbar_values();
    full_canvas_redraw();
}

void view_zoom_in(void)
{
    if (dstate.zoom < MAX_ZOOM) {
        dstate.zoom *= 2;
        handle_zoom_change();
    }
}

void view_zoom_out(void)
{
    if (dstate.zoom > MIN_ZOOM) {
        dstate.zoom /= 2;
        handle_zoom_change();
    }
}

void view_zoom_1(void)
{
    if (dstate.zoom != 1) {
        dstate.zoom = 1;
        handle_zoom_change();
    }
}

void view_zoom_2(void)
{
    if (dstate.zoom != 2) {
        dstate.zoom = 2;
        handle_zoom_change();
    }
}

void view_zoom_4(void)
{
    if (dstate.zoom != 4) {
        dstate.zoom = 4;
        handle_zoom_change();
    }
}

void view_zoom_8(void)
{
    if (dstate.zoom != 8) {
        dstate.zoom = 8;
        handle_zoom_change();
    }
}

void view_zoom_16(void)
{
    if (dstate.zoom != 16) {
        dstate.zoom = 16;
        handle_zoom_change();
    }
}

void view_fullscreen(void)
{
    dstate.fullscreen = !dstate.fullscreen;
    if (dstate.fullscreen) {
        gtk_widget_hide_all(gui.menubar);
        gtk_widget_hide_all(gui.toolbar_box);
        gtk_widget_hide_all(gui.sidebar);
        gtk_widget_hide(gui.vscrollbar);
        gtk_widget_hide(gui.hscrollbar);
    } else {
        gtk_widget_show_all(gui.menubar);
        gtk_widget_hide(gui.position_submenu);
        if (dstate.show_toolbar)
            gtk_widget_show_all(gui.toolbar_box);
        if (dstate.show_sidebar)
            gtk_widget_show_all(gui.sidebar);
        if (dstate.show_scrollbars) {
            gtk_widget_show(gui.vscrollbar);
            gtk_widget_show(gui.hscrollbar);
        }
    }
    ewmh_toggle_fullscreen(gui.window);
}

void view_show_toolbar(void)
{
    if (GTK_CHECK_MENU_ITEM(gui.show_toolbar_menu_item)->active == dstate.show_toolbar)
        return;

    dstate.show_toolbar = !dstate.show_toolbar;
    if (dstate.show_toolbar)
        gtk_widget_show_all(gui.toolbar_box);
    else
        gtk_widget_hide_all(gui.toolbar_box);
}

void view_show_sidebar(void)
{
    if (GTK_CHECK_MENU_ITEM(gui.show_sidebar_menu_item)->active == dstate.show_sidebar)
        return;

    dstate.show_sidebar = !dstate.show_sidebar;
    if (dstate.show_sidebar)
        gtk_widget_show_all(gui.sidebar);
    else
        gtk_widget_hide_all(gui.sidebar);
}

void view_show_scrollbars(void)
{
    if (GTK_CHECK_MENU_ITEM(gui.show_scrollbars_menu_item)->active == dstate.show_scrollbars)
        return;

    dstate.show_scrollbars = !dstate.show_scrollbars;
    if (dstate.show_scrollbars) {
        gtk_widget_show(gui.vscrollbar);
        gtk_widget_show(gui.hscrollbar);
    } else {
        gtk_widget_hide(gui.vscrollbar);
        gtk_widget_hide(gui.hscrollbar);
    }
}

void position_scroll_left(void)
{
    position_scroll_generic(-1, 0, SCROLL_STEP);
}

void position_scroll_right(void)
{
    position_scroll_generic(1, 0, SCROLL_STEP);
}

void position_scroll_up(void)
{
    position_scroll_generic(0, -1, SCROLL_STEP);
}

void position_scroll_down(void)
{
    position_scroll_generic(0, 1, SCROLL_STEP);
}

void position_scroll_nw(void)
{
    position_scroll_generic(-1, -1, SCROLL_STEP);
}

void position_scroll_ne(void)
{
    position_scroll_generic(1, -1, SCROLL_STEP);
}

void position_scroll_sw(void)
{
    position_scroll_generic(-1, 1, SCROLL_STEP);
}

void position_scroll_se(void)
{
    position_scroll_generic(1, 1, SCROLL_STEP);
}

void position_page_left(void)
{
    position_scroll_generic(-1, 0, SCROLL_PAGE);
}

void position_page_right(void)
{
    position_scroll_generic(1, 0, SCROLL_PAGE);
}

void position_page_up(void)
{
    position_scroll_generic(0, -1, SCROLL_PAGE);
}

void position_page_down(void)
{
    position_scroll_generic(0, 1, SCROLL_PAGE);
}

void position_page_nw(void)
{
    position_scroll_generic(-1, -1, SCROLL_PAGE);
}

void position_page_ne(void)
{
    position_scroll_generic(1, -1, SCROLL_PAGE);
}

void position_page_sw(void)
{
    position_scroll_generic(-1, 1, SCROLL_PAGE);
}

void position_page_se(void)
{
    position_scroll_generic(1, 1, SCROLL_PAGE);
}

void edit_cut(void)
{
    if (!selection_check())
        return;
    selection_copy_clear(TRUE, TRUE);
    state.selection = null_rect;
    full_canvas_redraw();
    update_population_label();
}

void edit_copy(void)
{
    if (!selection_check())
        return;
    selection_copy_clear(TRUE, FALSE);
    screen_box_update(&state.selection, &null_rect);
}

void edit_paste(void)
{
    point  pt;
    rect   r;

    if (state.tracking_mouse || !edit_paste_verify_buffer())
        return;
    state.tracking_mouse = PASTING;
    gtk_widget_get_pointer(gui.canvas, &pt.x, &pt.y);
    get_logical_coords(&pt, &r.start);
    r.end.x = r.start.x + (state.copy_rect.end.x - state.copy_rect.start.x);
    r.end.y = r.start.y + (state.copy_rect.end.y - state.copy_rect.start.y);
    if (r.end.x < WORLD_SIZE && r.end.y < WORLD_SIZE)
        screen_box_update(&state.paste_box, &r);
}

void edit_move(void)
{
    if (state.tracking_mouse || !selection_check())
        return;
    state.moving = TRUE;
    edit_paste();
}

void edit_clear(void)
{
    if (!selection_check())
        return;
    selection_copy_clear(FALSE, TRUE);
    state.selection = null_rect;
    full_canvas_redraw();
    update_population_label();
}

void edit_preferences(void)
{
    GtkWidget*  dialog;
    GtkWidget*  tabbed_frame;
    GtkWidget*  table;
    GtkWidget*  vbox;
    GtkWidget*  vbox2;
    GtkWidget*  hbox;
    GtkWidget*  frame;
    GtkWidget*  alignment;
    GtkWidget*  button;
    GtkWidget*  label;
    prefs_dialog_info*  info;
    color_dialog_info*  color_info;
    GList*      zoom_list = NULL;
    GList*      pos;
    char*       title_text;
    char*       label_text;
    char*       default_zoom_str;
    char*       buf;
    uint32      default_value;
    int32       i;

    /* Setup the dialog window */
    dialog = gtk_dialog_new();
    info = safe_malloc(sizeof(prefs_dialog_info));
    info->dialog = dialog;
    gtk_window_set_title(GTK_WINDOW(dialog), TITLE " Preferences");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", GTK_SIGNAL_FUNC(edit_preferences_destroy),
                       info);

    /* Init the tabbed frame */
    tabbed_frame = gtk_notebook_new();

    /*** Setup "File" preferences ***/

    frame = gtk_frame_new("General");
    gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
    table = gtk_table_new(1, 4, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 3);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    /* Patterns directory */
    LEFT_ALIGN(gtk_label_new("Patterns directory:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    info->patterns_dir_entry = gtk_entry_new_with_max_length(PATH_MAX);
    gtk_widget_set_usize(info->patterns_dir_entry, 200, 0);
    gtk_entry_set_text(GTK_ENTRY(info->patterns_dir_entry), config.patterns_dir);
    gtk_editable_set_position(GTK_EDITABLE(info->patterns_dir_entry), 0);
    LEFT_ALIGN(info->patterns_dir_entry);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    button = gtk_button_new_with_label("....");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(prefs_select_patterns_dir),
                       info);
    LEFT_ALIGN(button);
    gtk_table_attach(GTK_TABLE(table), alignment, 2, 3, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    button = gtk_button_new_with_label(" Default ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(prefs_patterns_dir),
                       info);
    LEFT_ALIGN(button);
    gtk_table_attach(GTK_TABLE(table), alignment, 3, 4, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Add "File" prefs to the tabbed frame */
    gtk_container_add(GTK_CONTAINER(frame), table);
    FRAME_ALIGN(frame);
    gtk_notebook_append_page(GTK_NOTEBOOK(tabbed_frame), alignment, gtk_label_new("File"));

    /*** Setup "View" preferences ***/

    vbox = gtk_vbox_new(FALSE, 0);

    /** "General" View Prefs **/

    frame = gtk_frame_new("General");
    gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
    vbox2 = gtk_vbox_new(FALSE, 10);
    gtk_container_set_border_width(GTK_CONTAINER(vbox2), 3);
    table = gtk_table_new(5, 2, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    /* Initial window size */
    LEFT_ALIGN(gtk_label_new("Initial window size:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    hbox = gtk_hbox_new(FALSE, 0);
    info->default_window_width_spinbox = create_spinbox(config.default_window_width, 0, 10000);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_window_width_spinbox, FALSE, FALSE, 0);
    label = gtk_label_new("X");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
    info->default_window_height_spinbox = create_spinbox(config.default_window_height, 0, 10000);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_window_height_spinbox, FALSE, FALSE, 0);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Default zoom level */
    LEFT_ALIGN(gtk_label_new("Default zoom level:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 1, 2, TABLE_OPTS, TABLE_OPTS, 0, 0);
    info->default_zoom_combo = gtk_combo_new();
    gtk_widget_set_usize(info->default_zoom_combo, 75, 0);
    default_zoom_str = "1:1";
    for (i=MIN_ZOOM; i <= MAX_ZOOM; i*=2) {
        buf = dsprintf("%d:1", i);
        zoom_list = g_list_append(zoom_list, buf);
        if (i == config.default_zoom)
            default_zoom_str = buf;
    }
    gtk_combo_set_popdown_strings(GTK_COMBO(info->default_zoom_combo), zoom_list);
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(info->default_zoom_combo)->entry), default_zoom_str);
    for (pos=zoom_list; pos; pos=g_list_next(pos))
        free(pos->data);
    g_list_free(zoom_list);
    LEFT_ALIGN(info->default_zoom_combo);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 1, 2, TABLE_OPTS, TABLE_OPTS, 0, 0);

    LEFT_ALIGN(gtk_label_new("Show by default:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 2, 3, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Default show toolbar */
    hbox = gtk_hbox_new(FALSE, 0);
    info->default_show_toolbar_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(info->default_show_toolbar_toggle),
                                 config.default_show_toolbar);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_show_toolbar_toggle, FALSE, FALSE, 0);
    label = gtk_label_new("Toolbar");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 2, 3, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Default show sidebar */
    hbox = gtk_hbox_new(FALSE, 0);
    info->default_show_sidebar_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(info->default_show_sidebar_toggle),
                                 config.default_show_sidebar);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_show_sidebar_toggle, FALSE, FALSE, 0);
    label = gtk_label_new("Sidebar");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 3, 4, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Default show scrollbars */
    hbox = gtk_hbox_new(FALSE, 0);
    info->default_show_scrollbars_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(info->default_show_scrollbars_toggle),
                                 config.default_show_scrollbars);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_show_scrollbars_toggle, FALSE, FALSE, 0);
    label = gtk_label_new("Scrollbars");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 4, 5, TABLE_OPTS, TABLE_OPTS, 0, 0);

    gtk_box_pack_start(GTK_BOX(vbox2), table, FALSE, FALSE, 0);

    /* Default fullscreen */
    hbox = gtk_hbox_new(FALSE, 0);
    info->default_fullscreen_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(info->default_fullscreen_toggle),
                                 config.default_fullscreen);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_fullscreen_toggle, FALSE, FALSE, 0);
    label = gtk_label_new("Start in fullscreen mode");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);

    /* Add "General" view prefs */
    gtk_container_add(GTK_CONTAINER(frame), vbox2);
    FRAME_ALIGN(frame);
    gtk_box_pack_start(GTK_BOX(vbox), alignment, FALSE, FALSE, 3);

    /** "Colors" View Prefs **/

    frame = gtk_frame_new("Colors");
    gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
    table = gtk_table_new(NUM_COLORS, 4, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 3);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    /* Loop through colors */
    for (i=0; i < NUM_COLORS; i++) {
        if (i == BG_COLOR_INDEX) {
            title_text    = "Choose Background Color";
            label_text    = "Background:";
            default_value = DEFAULT_BG_COLOR;
        } else if (i == CELL_COLOR_INDEX) {
            title_text  = "Choose Cell Color";
            label_text  = "Live cells:";
            default_value = DEFAULT_CELL_COLOR;
        } else if (i == GRID_COLOR_INDEX) {
            title_text  = "Choose Grid Color";
            label_text  = "Grid:";
            default_value = DEFAULT_GRID_COLOR;
        } else {
            title_text  = "Choose Selection Color";
            label_text  = "Selection:";
            default_value = DEFAULT_SELECT_COLOR;
        }
        color_info = info->color_dialog_infos[i] = safe_malloc(sizeof(color_dialog_info));

        LEFT_ALIGN(gtk_label_new(label_text));
        gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, i, i+1, TABLE_OPTS, TABLE_OPTS, 0, 0);
        info->color_entries[i] = gtk_entry_new();
        gtk_widget_set_usize(info->color_entries[i], 75, 0);
        buf = dsprintf("#%06X", config.colors[i]);
        gtk_entry_set_text(GTK_ENTRY(info->color_entries[i]), buf);
        free(buf);
        LEFT_ALIGN(info->color_entries[i]);
        gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, i, i+1, TABLE_OPTS, TABLE_OPTS, 0, 0);
        color_info->dialog_title  = title_text;
        color_info->prefs_entry   = info->color_entries[i];
        color_info->cur_value     = config.colors[i];
        color_info->default_value = default_value;
        button = gtk_button_new_with_label(" Select ");
        gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(prefs_select_color),
                           color_info);
        LEFT_ALIGN(button);
        gtk_table_attach(GTK_TABLE(table), alignment, 2, 3, i, i+1, TABLE_OPTS, TABLE_OPTS, 0, 0);

        button = gtk_button_new_with_label(" Default ");
        gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(prefs_default_color),
                           color_info);
        LEFT_ALIGN(button);
        gtk_table_attach(GTK_TABLE(table), alignment, 3, 4, i, i+1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    }

    /* Add "Colors" view prefs */
    gtk_container_add(GTK_CONTAINER(frame), table);
    FRAME_ALIGN(frame);
    gtk_box_pack_start(GTK_BOX(vbox), alignment, FALSE, FALSE, 3);

    /* Add "View" prefs to the tabbed frame */
    gtk_notebook_append_page(GTK_NOTEBOOK(tabbed_frame), vbox, gtk_label_new("View"));

    /*** Setup "Run" preferences ***/

    frame = gtk_frame_new("Speed");
    gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
    table = gtk_table_new(4, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 3);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    /* Default speed */
    LEFT_ALIGN(gtk_label_new("Default speed (Gen/s):"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    info->default_speed_spinbox = create_spinbox(config.default_speed, 1, 10000);
    LEFT_ALIGN(info->default_speed_spinbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Speed slider range */
    LEFT_ALIGN(gtk_label_new("Speed slider range:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 1, 2, TABLE_OPTS, TABLE_OPTS, 0, 0);
    hbox = gtk_hbox_new(FALSE, 0);
    info->speed_slider_min_spinbox = create_spinbox(config.speed_slider_min, 1, state.speed);
    gtk_box_pack_start(GTK_BOX(hbox), info->speed_slider_min_spinbox, FALSE, FALSE, 0);
    label = gtk_label_new("-");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
    info->speed_slider_max_spinbox = create_spinbox(config.speed_slider_max, state.speed, 10000);
    gtk_box_pack_start(GTK_BOX(hbox), info->speed_slider_max_spinbox, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 1, 2, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Speed increment */
    LEFT_ALIGN(gtk_label_new("Speed increment:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 2, 3, TABLE_OPTS, TABLE_OPTS, 0, 0);
    info->speed_increment_spinbox = create_spinbox(config.speed_increment, 1, 10000);
    LEFT_ALIGN(info->speed_increment_spinbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 2, 3, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Default skip frames */
    hbox = gtk_hbox_new(FALSE, 0);
    info->default_skip_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(info->default_skip_toggle),
                                 config.default_skip_frames);
    gtk_box_pack_start(GTK_BOX(hbox), info->default_skip_toggle, FALSE, FALSE, 0);
    label = gtk_label_new("Skip frames by default");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    LEFT_ALIGN(hbox);
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 2, 3, 4, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Add "Run" prefs to the tabbed frame */
    gtk_container_add(GTK_CONTAINER(frame), table);
    FRAME_ALIGN(frame);
    gtk_notebook_append_page(GTK_NOTEBOOK(tabbed_frame), alignment, gtk_label_new("Run"));

    /*** Setup "Help Browser" preferences ***/
    frame = gtk_frame_new("General");
    gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
    table = gtk_table_new(1, 3, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 3);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    /* Browser command-line */
    LEFT_ALIGN(gtk_label_new("Command line:"));
    gtk_table_attach(GTK_TABLE(table), alignment, 0, 1, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);
    info->help_browser_entry = gtk_entry_new();
    gtk_widget_set_usize(info->help_browser_entry, 250, 0);
    gtk_entry_set_text(GTK_ENTRY(info->help_browser_entry), config.help_browser);
    gtk_editable_set_position(GTK_EDITABLE(info->help_browser_entry), 0);
    LEFT_ALIGN(info->help_browser_entry);
    gtk_table_attach(GTK_TABLE(table), alignment, 1, 2, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    button = gtk_button_new_with_label(" Default ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(prefs_default_help_browser),
                       info);
    LEFT_ALIGN(button);
    gtk_table_attach(GTK_TABLE(table), alignment, 2, 3, 0, 1, TABLE_OPTS, TABLE_OPTS, 0, 0);

    /* Add "Help Browser" prefs to the tabbed frame */
    gtk_container_add(GTK_CONTAINER(frame), table);
    FRAME_ALIGN(frame);
    gtk_notebook_append_page(GTK_NOTEBOOK(tabbed_frame), alignment, gtk_label_new("Help Browser"));

    /* Add tabbed frame to dialog vbox */
    gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 10);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), tabbed_frame, FALSE, FALSE, 5);

    /* Setup the buttons */
    button  = gtk_button_new_with_label(" Okay ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(edit_preferences_ok), info);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(button);
    button  = gtk_button_new_with_label(" Apply ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(edit_preferences_apply),
                       info);
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);
    button = gtk_button_new_with_label(" Cancel ");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);

    /* Show the dialog */
    gtk_widget_show_all(dialog);
}

void run_start_stop(void)
{
    if (state.pattern_running && state.skipped_frames) {
        trigger_canvas_update();
        update_tick_label();
        update_population_label();
    }
    start_stop();
}

void run_step(void)
{
    if (!state.pattern_running)
        tick_and_update();
}

void run_faster(void)
{
    set_speed(state.speed + config.speed_increment);
}

void run_slower(void)
{
    set_speed(state.speed - config.speed_increment);
}

void run_speed(void)
{
    GtkWidget*  dialog;
    GtkWidget*  speed_spinbox;
    GtkWidget*  skip_toggle;
    GtkWidget*  button;
    GtkWidget*  label;
    GtkWidget*  hbox;
    speed_dialog_info*  info;

    info = safe_malloc(sizeof(speed_dialog_info));

    /* Setup the dialog window */
    dialog = gtk_dialog_new();
    gtk_window_set_title(GTK_WINDOW(dialog), "Set Speed");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", GTK_SIGNAL_FUNC(run_speed_destroy), info);

    /* Setup the Gen/s speed spinbox */
    hbox = gtk_hbox_new(FALSE, 0);
    label = gtk_label_new("Speed in generations/sec:");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
    speed_spinbox = create_spinbox(state.speed, 1, 10000);
    gtk_box_pack_start(GTK_BOX(hbox), speed_spinbox, FALSE, FALSE, 5);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox, FALSE, FALSE, 5);

    /* Setup the "skip frames" toggle */
    hbox = gtk_hbox_new(FALSE, 0);
    skip_toggle = gtk_check_button_new();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(skip_toggle), state.skip_frames);
    gtk_box_pack_start(GTK_BOX(hbox), skip_toggle, FALSE, FALSE, 3);
    label = gtk_label_new("Skip frames to achieve speed");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox, FALSE, FALSE, 5);

    /* Data to be sent to "okay" button signal handler */
    info->dialog        = dialog;
    info->speed_spinbox = speed_spinbox;
    info->skip_toggle   = skip_toggle;

    /* Setup the buttons */
    button  = gtk_button_new_with_label(" Okay ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(run_speed_ok), info);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
    button = gtk_button_new_with_label(" Apply ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(run_speed_apply), info);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
    button = gtk_button_new_with_label(" Cancel ");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);

    /* Put keyboard focus on the speed spinbox */
    gtk_widget_grab_focus(speed_spinbox);

    /* Show the dialog */
    gtk_widget_show_all(dialog);
}

void help_help(void)
{
    char*  url = "file://" DOCDIR "/index.html";
    char*  command;
    char*  old_command;
    char*  p;

    command = dsprintf("%s &", config.help_browser);
    while ((p = strstr(command, HELP_BROWSER_PARAM)) != NULL) {
        *p = '\0';
        old_command = command;
        command = dsprintf("%s%s%s", command, url, p + strlen(HELP_BROWSER_PARAM));
        free(old_command);
    }
    system(command);
    free(command);
}

void help_about(void)
{
    GtkWidget*  dialog;
    GtkWidget*  ok_button;
    GtkWidget*  vbox;
    GtkWidget*  hbox;
    GtkWidget*  alignment;
    GtkWidget*  pixmap;

    /* Set up the dialog window */
    dialog = gtk_dialog_new();
    gtk_window_set_title(GTK_WINDOW(dialog), "About " TITLE);
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);

    /* Add the banner */
    pixmap = load_pixmap_from_xpm_file(DATADIR "/graphics/banner.xpm");
    if (pixmap)
        gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), pixmap, FALSE, FALSE, 5);

    hbox = gtk_hbox_new(FALSE, 0);

    /* Add the logo */
    pixmap = load_pixmap_from_xpm_file(DATADIR "/graphics/logo.xpm");
    if (pixmap)
        gtk_box_pack_start(GTK_BOX(hbox), pixmap, FALSE, FALSE, 5);

    /* Add the text */
    vbox = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new(TITLE " Version " VERSION), FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new("Suzanne Skinner"), FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new("Copyright 2004"), FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new("GNU General Public License"), FALSE, FALSE,
                       0);
    CENTER_ALIGN(vbox);
    gtk_box_pack_start(GTK_BOX(hbox), alignment, TRUE, TRUE, 0);

    /* Add the "okay" button */
    ok_button = gtk_button_new_with_label(" Okay ");
    gtk_signal_connect_object(GTK_OBJECT(ok_button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), ok_button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(ok_button);

    /* Show the dialog */
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox, FALSE, FALSE, 5);
    gtk_widget_show_all(dialog);
}

/*** Bound Function Helpers ***/

/* A helper for the various scroll functions. xdir indicates how to scroll along the x axis (-1 for
 * left, 0 for none, 1 for right), ydir indicates the y axis scrolling, and how_far indicates how
 * much to scroll (this parameter may take the special values SCROLL_STEP and SCROLL_PAGE, to
 * use the scrollbar step and page increments). This function just adjusts the scrollbar(s), and
 * lets the scrollbar event handlers take care of the rest.
 */
void position_scroll_generic(int32 xdir, int32 ydir, int32 how_far)
{
    GtkWidget*      scrollbar;
    GtkAdjustment*  sb_state;
    int32           dir;
    int32           i;

    for (i=0; i < 2; i++) {
        if (i == 0) {
            dir = xdir;
            scrollbar = gui.hscrollbar;
        } else {
            dir = ydir;
            scrollbar = gui.vscrollbar;
        }
        if (!dir)
            continue;
        sb_state = gtk_range_get_adjustment(GTK_RANGE(scrollbar));
        if (how_far == SCROLL_STEP)
            how_far = sb_state->step_increment;
        else if (how_far == SCROLL_PAGE)
            how_far = sb_state->page_increment;
        gtk_adjustment_set_value(sb_state, ROUND(sb_state->value) + dir * how_far);
    }
}

/* Called when the user selects a file from a file open dialog */
void file_open_ok(GtkWidget* widget, GtkFileSelection* file_selector)
{
    char*  path;

    path = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector));
    set_new_current_dir(path);
    if (attempt_load_pattern(path)) {
        sidebar_unselect();
        gtk_widget_destroy(GTK_WIDGET(file_selector));
    }
}

/* Called when the user selects a file from a "save as" dialog */
void file_save_as_ok(GtkWidget* widget, GtkFileSelection* file_selector)
{
    struct stat  statbuf;
    char*  path;

    path = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector));
    if (stat(path, &statbuf) == 0)   /* file exists */
        file_save_as_confirm_overwrite(widget, file_selector);
    else {
        gtk_object_set_user_data(GTK_OBJECT(file_selector), NULL);
        file_save_as_confirmed(widget, file_selector);
    }
}

/* Called to generate an "are you sure" dialog when the user is about to overwrite
 * an existing file.
 */
void file_save_as_confirm_overwrite(GtkWidget* widget, GtkFileSelection* file_selector) {
    GtkWidget*  dialog;
    GtkWidget*  label;
    GtkWidget*  button;
    GtkWidget*  alignment;
    char*       path;
    char*       message;

    dialog = gtk_dialog_new();
    gtk_window_set_title(GTK_WINDOW(dialog), "File Exists");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
    gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
    gtk_object_set_user_data(GTK_OBJECT(file_selector), dialog);
    gtk_container_set_border_width(GTK_CONTAINER(dialog), 0);
    gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 5);

    /* Set up the confirmation question */
    path = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector));
    message = dsprintf("%s exists, overwrite?", path);
    label = gtk_label_new(message);
    free(message);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), label);

    /* Set up the "okay" and "cancel" buttons */
    button = gtk_button_new_with_label(" Okay ");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(file_save_as_confirmed),
                       file_selector);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(button);
    button = gtk_button_new_with_label(" Cancel ");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
    CENTER_ALIGN(button);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), alignment, FALSE, FALSE, 0);

    /* Show the dialog */
    gtk_widget_show_all(dialog);
}

/* Called when it's time to actually save to a file--any user confirmation is done. */
void file_save_as_confirmed(GtkWidget* widget, GtkFileSelection* file_selector)
{
    GtkWidget*  confirm_dialog;
    char*       path;

    confirm_dialog = gtk_object_get_user_data(GTK_OBJECT(file_selector));
    if (confirm_dialog)
        gtk_widget_destroy(confirm_dialog);
    path = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector));
    set_new_current_dir(path);
    if (attempt_save_pattern(path)) {
        sidebar_unselect();
        gtk_widget_destroy(GTK_WIDGET(file_selector));
    }
}

/* Called when the user clicks "Okay" from the pattern description dialog */
void file_description_ok(GtkWidget* widget, desc_dialog_info* info)
{
    if (file_description_apply(widget, info))
        gtk_widget_destroy(info->dialog);
}

/* Called when the user clicks "Apply" from the pattern description dialog. Returns TRUE
 * if the description change was successful.
 */
boolean file_description_apply(GtkWidget* widget, desc_dialog_info* info)
{
    static char  buf[DESCRIPTION_MAX_LINES][DESCRIPTION_MAX_COLS+1];
    int32    num_lines = 0;
    boolean  failed = FALSE;
    char*    desc;
    char*    pos;
    char*    newline_pos;
    int32    i;

    desc = gtk_editable_get_chars(GTK_EDITABLE(info->textbox), 0, -1);
    pos = desc;
    while (pos && *pos && !failed) {
        if (num_lines == DESCRIPTION_MAX_LINES)
            failed = TRUE;
        else {
            newline_pos = strchr(pos, '\n');
            if (newline_pos)
                *newline_pos = '\0';
            if (strlen(pos) > DESCRIPTION_MAX_COLS)
                failed = TRUE;
            else {
                strcpy(buf[num_lines++], pos);
                pos = (newline_pos ? newline_pos+1 : NULL);
            }
        }
    }

    free(desc);
    if (failed) {
        error_dialog("The description for a .lif file must be no more\n"
                     "than %d lines long and %d characters wide.\n"
                     "Please fix and retry.",
                     DESCRIPTION_MAX_LINES, DESCRIPTION_MAX_COLS);
        return FALSE;
    } else {
        for (i=0; i < num_lines; i++)
            strcpy(pattern_description[i], buf[i]);
        desc_num_lines = num_lines;
        return TRUE;
    }
}

/* Called when the description dialog is closed */
void file_description_destroy(GtkWidget* widget, desc_dialog_info* info)
{
    free(info);
}

/* Called when the user chooses where they want to paste after typing "ctrl-v".
 */
void edit_paste_win_style(const point* pt)
{
    if (!edit_paste_verify_target(pt))
        return;
    if (state.moving) {
        selection_copy_clear(FALSE, TRUE);
        state.moving = FALSE;
    }
    edit_paste_perform(pt);
    state.tracking_mouse = NOT_TRACKING;
    state.selection = state.paste_box = null_rect;
    full_canvas_redraw();
    update_population_label();
}

/* Called when the user middle-clicks on a cell to paste.
 */
void edit_paste_unix_style(const point* pt)
{
    if (state.tracking_mouse || !edit_paste_verify_buffer() || !edit_paste_verify_target(pt))
        return;
    edit_paste_perform(pt);
    state.selection = null_rect;
    full_canvas_redraw();
    update_population_label();
}

/* Verify that there is either a currently selected region or a copy buffer to paste from,
 * displaying an error dialog and returning FALSE if not. If there is a currently selected
 * region, implicitly copy it.
 */
boolean edit_paste_verify_buffer(void)
{
    if (SELECTION_ACTIVE())
        selection_copy_clear(TRUE, FALSE);
    else if (!state.copy_buffer) {
        error_dialog("There is no active selection or copy buffer.");
        return FALSE;
    }
    return TRUE;
}

/* Verify that the current copy buffer can be pasted to the given point (i.e., it won't go off
 * the end of the world). Display an error dialog and return FALSE if not, otherwise return TRUE.
 */
boolean edit_paste_verify_target(const point* pt)
{
    if (pt->x + (state.copy_rect.end.x - state.copy_rect.start.x) < WORLD_SIZE &&
        pt->y + (state.copy_rect.end.y - state.copy_rect.start.y) < WORLD_SIZE)
        return TRUE;
    else {
        error_dialog("There's not enough room to paste there.");
        return FALSE;
    }
}

/* Paste the current copy buffer at the given location.
 */
void edit_paste_perform(const point* pt)
{
    int32       xstart, ystart;
    int32       xoff, yoff;
    cage_type*  c;

    for (c=state.copy_buffer; c; c=c->next) {
        xstart = pt->x + (c->x * CAGE_SIZE) - state.copy_rect.start.x;
        ystart = pt->y + (c->y * CAGE_SIZE) - state.copy_rect.start.y;
        for (yoff=0; yoff < 4; yoff++) {
            for (xoff=0; xoff < 4; xoff++) {
                if (c->bnw[0] & (1 << (yoff*4 + xoff)))
                    draw_cell(xstart+xoff, ystart+yoff, DRAW_SET);
            }
        }
        for (yoff=0; yoff < 4; yoff++) {
            for (xoff=0; xoff < 4; xoff++) {
                if (c->bne[0] & (1 << (yoff*4 + xoff)))
                    draw_cell(xstart+4+xoff, ystart+yoff, DRAW_SET);
            }
        }
        for (yoff=0; yoff < 4; yoff++) {
            for (xoff=0; xoff < 4; xoff++) {
                if (c->bsw[0] & (1 << (yoff*4 + xoff)))
                    draw_cell(xstart+xoff, ystart+4+yoff, DRAW_SET);
            }
        }
        for (yoff=0; yoff < 4; yoff++) {
            for (xoff=0; xoff < 4; xoff++) {
                if (c->bse[0] & (1 << (yoff*4 + xoff)))
                    draw_cell(xstart+4+xoff, ystart+4+yoff, DRAW_SET);
            }
        }
    }
}

/* Called when the user clicks "Okay" in the preferences dialog */
void edit_preferences_ok(GtkWidget* widget, prefs_dialog_info* info)
{
    edit_preferences_apply(widget, info);
    gtk_widget_destroy(info->dialog);
}

/* Called when the user clicks "Apply" in the preferences dialog */
void edit_preferences_apply(GtkWidget* widget, prefs_dialog_info* info)
{
    static int32 powers_of_2[] = {1, 2, 4, 8, 16, 32, 64, 128, 256, 0};
    GtkAdjustment*  slider_state;
    char*           color_str;
    int32           i;

    /* Apply "File" preferences */
    free(config.patterns_dir);
    config.patterns_dir = safe_strdup(gtk_entry_get_text(GTK_ENTRY(info->patterns_dir_entry)));
    if (config.patterns_dir[strlen(config.patterns_dir)-1] != '/')
        dstrcat(&config.patterns_dir, "/");
    update_pattern_list();

    /* Apply General "View" preferences */
    config.default_window_width =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->default_window_width_spinbox));
    config.default_window_height =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->default_window_height_spinbox));
    config.default_zoom =
        atoi(gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(info->default_zoom_combo)->entry)));
    if (config.default_zoom < MIN_ZOOM)
        config.default_zoom = MIN_ZOOM;
    else if (config.default_zoom > MAX_ZOOM)
        config.default_zoom = MAX_ZOOM;
    else {
        for (i=0; powers_of_2[i]; i++) {
            if (powers_of_2[i] >= config.default_zoom) {
                config.default_zoom = powers_of_2[i];
                break;
            }
        }
    }
    config.default_show_toolbar = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(info->default_show_toolbar_toggle));
    config.default_show_sidebar = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(info->default_show_sidebar_toggle));
    config.default_show_scrollbars = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(info->default_show_scrollbars_toggle));
    config.default_fullscreen = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(info->default_fullscreen_toggle));

    /* Apply Color "View" preferences */
    for (i=0; i < NUM_COLORS; i++) {
        color_str = gtk_entry_get_text(GTK_ENTRY(info->color_entries[i]));
        if (color_str[0] == '#')
            color_str++;
        config.colors[i] = MIN(0xFFFFFF, strtoul(color_str, NULL, 16));
    }
    dstate.life_pixmap_colormap = gdk_rgb_cmap_new(config.colors, NUM_COLORS);

    /* Apply "Run" preferences */
    config.default_speed =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->default_speed_spinbox));
    config.speed_slider_min =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->speed_slider_min_spinbox));;
    if (config.speed_slider_min > config.default_speed)
        config.speed_slider_min = config.default_speed;
    config.speed_slider_max =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->speed_slider_max_spinbox));;
    if (config.speed_slider_max < config.default_speed)
        config.speed_slider_max = config.default_speed;
    set_speed_label_size(config.speed_slider_max);
    config.speed_increment =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->speed_increment_spinbox));;
    slider_state = gtk_range_get_adjustment(GTK_RANGE(gui.speed_slider));
    slider_state->lower = config.speed_slider_min;
    slider_state->upper = config.speed_slider_max;
    slider_state->page_increment = config.speed_increment;
    gtk_adjustment_changed(slider_state);
    config.default_skip_frames = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(info->default_skip_toggle));

    /* Apply "Help Browser" preferences */
    free(config.help_browser);
    config.help_browser = safe_strdup(gtk_entry_get_text(GTK_ENTRY(info->help_browser_entry)));

    save_preferences();
    full_canvas_redraw();
}

/* Called when the preferences dialog is closed */
void edit_preferences_destroy(GtkWidget* widget, prefs_dialog_info* info)
{
    int32  i;

    for (i=0; i < NUM_COLORS; i++)
        free(info->color_dialog_infos[i]);
    free(info);
}

/* Called when the user clicks "Okay" from the set speed dialog.
 */
void run_speed_ok(GtkWidget* widget, speed_dialog_info* info)
{
    run_speed_apply(widget, info);
    gtk_widget_destroy(info->dialog);
}

/* called when the "Apply" button is clicked in the set speed dialog */
void run_speed_apply(GtkWidget* widget, speed_dialog_info* info)
{
    set_speed(gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(info->speed_spinbox)));
    state.skip_frames = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(info->skip_toggle));
}

/* Called when the speed dialog is closed */
void run_speed_destroy(GtkWidget* widget, speed_dialog_info* info)
{
    free(info);
}

/* Called when the user clicks the "..." button to select a new patterns dir in preferences */
void prefs_select_patterns_dir(GtkWidget* widget, prefs_dialog_info* info)
{
    GtkWidget*  file_selector;
    char*       dir;

    /* Setup the file selector */
    file_selector = gtk_file_selection_new("Choose Patterns Directory");
    gtk_window_set_modal(GTK_WINDOW(file_selector), TRUE);
    gtk_window_set_position(GTK_WINDOW(file_selector), GTK_WIN_POS_CENTER);
    gtk_object_set_user_data(GTK_OBJECT(file_selector), info);
    dir = safe_strdup(gtk_entry_get_text(GTK_ENTRY(info->patterns_dir_entry)));
    if (dir[strlen(dir)-1] != '/')
        dstrcat(&dir, "/");
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_selector), dir);
    free(dir);

    /* Link the buttons to handlers */
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->ok_button), "clicked",
                       GTK_SIGNAL_FUNC(prefs_select_patterns_dir_ok), file_selector);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->cancel_button),
                              "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(file_selector));

    /* Show the dialog */
    gtk_widget_show(file_selector);
}

/* Called when the user finishes selecting a new patterns dir */
void prefs_select_patterns_dir_ok(GtkWidget* widget, GtkFileSelection* file_selector)
{
    prefs_dialog_info*  info;

    info = gtk_object_get_user_data(GTK_OBJECT(file_selector));
    gtk_entry_set_text(GTK_ENTRY(info->patterns_dir_entry),
                       gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector)));
    gtk_widget_destroy(GTK_WIDGET(file_selector));
}

/* Called when the user clicks the "default" button for patterns directory in preferences */
void prefs_patterns_dir(GtkWidget* widget, prefs_dialog_info* info)
{
    gtk_entry_set_text(GTK_ENTRY(info->patterns_dir_entry), DEFAULT_PATTERNS_DIR);
}

/* Called when the user clicks a "select" button to change a color in preferences */
void prefs_select_color(GtkWidget* widget, color_dialog_info* info)
{
    GtkWidget*  color_dialog;
    double      color_array[4];
    int32       i;

    /* Setup the color selector */
    color_dialog = gtk_color_selection_dialog_new(info->dialog_title);
    gtk_window_set_modal(GTK_WINDOW(color_dialog), TRUE);
    gtk_window_set_position(GTK_WINDOW(color_dialog), GTK_WIN_POS_CENTER);
    gtk_object_set_user_data(GTK_OBJECT(color_dialog), info);
    color_array[0] = (double)(info->cur_value >> 16)         / 255.0;
    color_array[1] = (double)((info->cur_value >> 8) & 0xFF) / 255.0;
    color_array[2] = (double)(info->cur_value & 0xFF)        / 255.0;
    for (i=0; i < 2; i++)
        gtk_color_selection_set_color(
            GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(color_dialog)->colorsel), color_array);

    /* Link the buttons to handlers */
    gtk_signal_connect(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(color_dialog)->ok_button), "clicked",
                       GTK_SIGNAL_FUNC(prefs_select_color_ok), color_dialog);
    gtk_signal_connect_object(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(color_dialog)->cancel_button),
                              "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(color_dialog));

    /* Show the dialog */
    gtk_widget_hide(GTK_COLOR_SELECTION_DIALOG(color_dialog)->help_button);
    gtk_widget_show(color_dialog);
}

/* Called when the user finishes selecting a new color in preferences */
void prefs_select_color_ok(GtkWidget* widget, GtkColorSelectionDialog* color_dialog)
{
    color_dialog_info*  info;
    double              color_array[4];
    char*               color_str;

    info = gtk_object_get_user_data(GTK_OBJECT(color_dialog));
    gtk_color_selection_get_color(
        GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(color_dialog)->colorsel), color_array);
    info->cur_value = ((int)(color_array[0] * 255.0) << 16) |
                      ((int)(color_array[1] * 255.0) << 8)  |
                      ((int)(color_array[2] * 255.0));
    color_str = dsprintf("#%06X", info->cur_value);
    gtk_entry_set_text(GTK_ENTRY(info->prefs_entry), color_str);
    free(color_str);
    gtk_widget_destroy(GTK_WIDGET(color_dialog));
}

/* Called when the user clicks "default" to reset a color in preferences */
void prefs_default_color(GtkWidget* widget, color_dialog_info* info)
{
    char*  color_str;

    info->cur_value = info->default_value;
    color_str = dsprintf("#%06X", info->cur_value);
    gtk_entry_set_text(GTK_ENTRY(info->prefs_entry), color_str);
    free(color_str);
}

/* Called when the user clicks "default" to reset the help browser in preferences */
void prefs_default_help_browser(GtkWidget* widget, prefs_dialog_info* info)
{
    gtk_entry_set_text(GTK_ENTRY(info->help_browser_entry), DEFAULT_HELP_BROWSER);
}

/*** Miscellaneous Functions ***/

/* Add the current file to the top of the recent files list, and update the File menu accordingly.
 * If it is already in the list, move it to the top.
 */
void add_to_recent_files(void)
{
    char     full_path[PATH_MAX];
    char*    filename;
    char*    slash_pos;
    char*    label_text;
    boolean  maxed_out = FALSE;
    int32    num_recent;
    int32    old_pos;
    int32    start;
    int32    i;

    /* Get the full path of the current file */
    if (!realpath(state.pattern_path, full_path)) {
        sys_warn("can't get realpath of %s", state.pattern_path);
        strcpy(full_path, state.pattern_path);
    }

    /* Is it already at the top of recent files? */
    if (state.recent_files[0].full_path && STR_EQUAL(full_path, state.recent_files[0].full_path))
        return;

    /* Determine filename from full_path */
    slash_pos = strrchr(full_path, '/');
    filename = safe_strdup(slash_pos ? slash_pos+1 : full_path);

    /* Determine the current # of recent files, and the position of this one if it's already in
     * the list.
     */
    for (i=0, old_pos=0; i < MAX_RECENT_FILES && state.recent_files[i].full_path; i++) {
        if (STR_EQUAL(state.recent_files[i].full_path, full_path))
            old_pos = i;
    }
    num_recent = i;

    /* If the recent files list is at maximum and this is a new entry, drop off the last item */
    if (num_recent == MAX_RECENT_FILES && !old_pos) {
        maxed_out = TRUE;
        num_recent--;
    }

    /* Make room for the new item at the top of the list */
    start = (old_pos ? old_pos : num_recent);
    free(state.recent_files[start].full_path);
    free(state.recent_files[start].filename);
    for (i=start; i > 0; i--) {
        state.recent_files[i].full_path = state.recent_files[i-1].full_path;
        state.recent_files[i].filename  = state.recent_files[i-1].filename;
        label_text = dsprintf("%d. %s", i+1, state.recent_files[i].filename);
        gtk_label_set_text(GTK_LABEL(state.recent_files[i].label), label_text);
        free(label_text);
    }

    /* Put the new item at the top */
    state.recent_files[0].full_path = safe_strdup(full_path);
    state.recent_files[0].filename  = filename;
    label_text = dsprintf("1. %s", filename);
    gtk_label_set_text(GTK_LABEL(state.recent_files[i].label), label_text);
    free(label_text);

    /* If we've increased the length of the list, show the menu item which was hidden before */
    if (!maxed_out && !old_pos)
        gtk_widget_show(state.recent_files[num_recent].menu_item);
    /* If the list was empty before this, show the menu separator which was hidden before */
    if (!num_recent)
        gtk_widget_show(gui.recent_files_separator);
}

/* Attempt to load a pattern file from the given path, displaying an error dialog and returning
 * FALSE on failure, otherwise returning TRUE. If the original path cannot be loaded, also try
 * with .DEFAULT_EXTENSION appended. If the load was successful, this function will stop any
 * running pattern, update the tick label, and recenter and redraw the canvas.
 */
boolean attempt_load_pattern(const char* path)
{
    char*             path2;
    boolean           with_extension = FALSE;
    load_result_type  load_result;

    load_result = load_pattern(path);
    if (load_result != LOAD_SUCCESS) {
        path2 = dsprintf("%s.%s", path, DEFAULT_EXTENSION);
        if (load_pattern(path2) == LOAD_SUCCESS) {
            path = path2;
            with_extension = TRUE;
            load_result = LOAD_SUCCESS;
        }
        else
            free(path2);
    }

    if (load_result == LOAD_SUCCESS) {
        if (!state.pattern_path || !STR_EQUAL(path, state.pattern_path)) {
            free(state.pattern_path);
            state.pattern_path = (with_extension ? (char*)path : safe_strdup(path));
            add_to_recent_files();
            put_filename_in_window_title();
        }
        if (state.pattern_running)
            start_stop();
        state.tracking_mouse = NOT_TRACKING;
        state.selection = state.paste_box = null_rect;
        state.last_drawn = null_point;
        state.moving = FALSE;
        update_tick_label();
        update_population_label();
        view_recenter();   /* will trigger a canvas redraw */
        return TRUE;
    }

    if (load_result == LOAD_NOT_FOUND)
        error_dialog("Open failed: \"%s\" not found", path);
    else if (load_result == LOAD_SYS_ERROR)
        error_dialog("Open failed: %s", strerror(errno));
    else
        error_dialog("Open failed: invalid format in \"%s\"", path);
    return FALSE;
}

/* Attempt to save a pattern file to the given path, displaying an error dialog and returning
 * FALSE on failure, otherwise returning TRUE.
 */
boolean attempt_save_pattern(const char* path)
{
    if (save_pattern(path)) {
        if (path != state.pattern_path) {
            free(state.pattern_path);
            state.pattern_path = safe_strdup(path);
            add_to_recent_files();
            put_filename_in_window_title();
        }
        return TRUE;
    }
    else {
        error_dialog("Save failed: %s", strerror(errno));
        return FALSE;
    }
}

/* Determine the Life coordinates (from 0 to 65535) that correspond to the given screen
 * coordinates (relative to the canvas). If the given coordinates are out of bounds, use the
 * nearest point on the canvas.
 */
void get_logical_coords(const point* p, point* logical_p)
{
    point  pt;
    int32  grid_block_size;

    pt = *p;
    if (pt.x < 0)
        pt.x = 0;
    else if (pt.x >= dstate.eff_canvas_size.width)
        pt.x = dstate.eff_canvas_size.width - 1;
    if (pt.y < 0)
        pt.y = 0;
    else if (pt.y >= dstate.eff_canvas_size.height)
        pt.y = dstate.eff_canvas_size.height - 1;

    grid_block_size = (DRAW_GRID(dstate.zoom) ? dstate.zoom+1 : dstate.zoom);
    logical_p->x = MIN(dstate.viewport.end.x, dstate.viewport.start.x + pt.x / grid_block_size);
    logical_p->y = MIN(dstate.viewport.end.y, dstate.viewport.start.y + pt.y / grid_block_size);
}

/* Determine the screen coordinates that correspond to the given Life coordinates. Specifically,
 * determine the location of the upper left hand pixel of the onscreen block (past any grid
 * lines).
 */
void get_screen_coords(const point* p, point* screen_p)
{
    int32  grid_block_size;

    grid_block_size = (DRAW_GRID(dstate.zoom) ? dstate.zoom+1 : dstate.zoom);
    screen_p->x = (p->x - dstate.viewport.start.x) * grid_block_size;
    screen_p->y = (p->y - dstate.viewport.start.y) * grid_block_size;
    if (DRAW_GRID(dstate.zoom)) {
        screen_p->x++;
        screen_p->y++;
    }
}

/* Call get_screen_coords to get screen coordinates for the Life coordinates of the given
 * rectangle.
 */
void get_screen_rectangle(const rect* r, rect* screen_r)
{
    get_screen_coords(&(r->start), &(screen_r->start));
    get_screen_coords(&(r->end),   &(screen_r->end));
}

/* Process an update of an onscreen rectangle (selection or paste): set the new values, then redraw
 * the offscreen pixmap and invalidate portions of the screen to account for the visible change.
 * Pass &null_rect for newr to indicate that the rectangle has been deactivated.
 */
void screen_box_update(rect* oldr, const rect* newr)
{
    rect       r, or, sr, cr;
    dimension  dim;
    int32      i;

    if (rects_identical(oldr, newr))
        return;
    or = *oldr;
    *oldr = *newr;

    if (DRAW_GRID(dstate.zoom))
        draw_grid_and_boxes();
    else
        draw_life_pixmap();

    for (i=0; i < 2; i++) {
        if (i == 0) {
            if (or.start.x < 0)
                continue;
            r = or;
        } else {
            if (newr->start.x < 0)
                continue;
            r = *newr;
        }

        normalize_rectangle(&r);
        get_screen_rectangle(&r, &sr);
        sr.end.x += dstate.zoom-1;
        sr.end.y += dstate.zoom-1;
        if (DRAW_GRID(dstate.zoom)) {
            sr.start.x--;
            sr.start.y--;
            sr.end.x++;
            sr.end.y++;
        }
        cr.start.x = MAX(0, sr.start.x);
        cr.start.y = MAX(0, sr.start.y);
        cr.end.x   = MIN(dstate.eff_canvas_size.width-1, sr.end.x);
        cr.end.y   = MIN(dstate.eff_canvas_size.height-1, sr.end.y);
        get_rect_dimensions(&cr, &dim);

        if (sr.start.x >= 0)
            gtk_widget_queue_draw_area(gui.canvas, sr.start.x, cr.start.y, 1, dim.height);
        if (sr.end.x < dstate.eff_canvas_size.width)
            gtk_widget_queue_draw_area(gui.canvas, sr.end.x, cr.start.y, 1, dim.height);
        if (sr.start.y >= 0)
            gtk_widget_queue_draw_area(gui.canvas, cr.start.x, sr.start.y, dim.width, 1);
        if (sr.end.y < dstate.eff_canvas_size.height)
            gtk_widget_queue_draw_area(gui.canvas, cr.start.x, sr.end.y, dim.width, 1);
    }
}

boolean selection_check(void)
{
    if (SELECTION_ACTIVE())
        return TRUE;
    else {
        error_dialog("No active selection: shift-click and drag to create one");
        return FALSE;
    }
}

/* Clear and/or copy the current selection. */
void selection_copy_clear(boolean copy, boolean clear)
{
    static uint16 left_col_masks[8]   = {0xFFFF, 0xEEEE, 0xCCCC, 0x8888, 0, 0, 0, 0};
    static uint16 right_col_masks[8]  = {0xFFFF, 0x7777, 0x3333, 0x1111, 0, 0, 0, 0};
    static uint16 top_col_masks[8]    = {0xFFFF, 0xFFF0, 0xFF00, 0xF000, 0, 0, 0, 0};
    static uint16 bottom_col_masks[8] = {0xFFFF, 0x0FFF, 0x00FF, 0x000F, 0, 0, 0, 0};
    cage_type*  c;
    cage_type*  newcage;
    rect        r;
    int32       left_offset, right_offset, top_offset, bottom_offset;
    uint16      nw_mask, ne_mask, sw_mask, se_mask;
    int32       x, y;

    r = state.selection;
    normalize_rectangle(&r);
    if (copy) {
        state.copy_rect = r;
        clear_cage_list(&state.copy_buffer);
    }

    while ((c = loop_cages()) != NULL) {
        if (IS_EMPTY(c))
            continue;
        x = c->x * CAGE_SIZE;
        y = c->y * CAGE_SIZE;
        if (r.end.x >= x && r.start.x < x+CAGE_SIZE &&
            r.end.y >= y && r.start.y < y+CAGE_SIZE) {
            nw_mask = ne_mask = sw_mask = se_mask = 0xFFFF;
            if (r.start.x > x || r.start.y > y ||
                r.end.x < x+CAGE_SIZE-1 || r.end.y < y+CAGE_SIZE-1) {

                left_offset   = MAX(0, r.start.x - x);
                right_offset  = MAX(0, x + CAGE_SIZE - 1 - r.end.x);
                top_offset    = MAX(0, r.start.y - y);
                bottom_offset = MAX(0, y + CAGE_SIZE - 1 - r.end.y);

                nw_mask &= left_col_masks[left_offset];
                sw_mask &= left_col_masks[left_offset];
                ne_mask &= left_col_masks[MAX(left_offset-4, 0)];
                se_mask &= left_col_masks[MAX(left_offset-4, 0)];
                ne_mask &= right_col_masks[right_offset];
                se_mask &= right_col_masks[right_offset];
                nw_mask &= right_col_masks[MAX(right_offset-4, 0)];
                sw_mask &= right_col_masks[MAX(right_offset-4, 0)];
                nw_mask &= top_col_masks[top_offset];
                ne_mask &= top_col_masks[top_offset];
                sw_mask &= top_col_masks[MAX(top_offset-4, 0)];
                se_mask &= top_col_masks[MAX(top_offset-4, 0)];
                sw_mask &= bottom_col_masks[bottom_offset];
                se_mask &= bottom_col_masks[bottom_offset];
                nw_mask &= bottom_col_masks[MAX(bottom_offset-4, 0)];
                ne_mask &= bottom_col_masks[MAX(bottom_offset-4, 0)];
            }
            if (copy) {
                newcage = safe_malloc(sizeof(cage_type));
                newcage->x = c->x;
                newcage->y = c->y;
                newcage->bnw[0] = c->bnw[parity] & nw_mask;
                newcage->bne[0] = c->bne[parity] & ne_mask;
                newcage->bsw[0] = c->bsw[parity] & sw_mask;
                newcage->bse[0] = c->bse[parity] & se_mask;
                newcage->next = state.copy_buffer;
                state.copy_buffer = newcage;
            }
            if (clear)
                mask_cage(c, ~nw_mask, ~ne_mask, ~sw_mask, ~se_mask);
        }
    }
}

/* Set state.current_dir (default load directory) to the parent directory of the given file.
 */
void set_new_current_dir(const char* filepath)
{
    char*  parent_dir;
    char*  slash_pos;

    parent_dir = safe_strdup(filepath);
    slash_pos = strrchr(parent_dir, '/');
    if (!slash_pos) {
        free(parent_dir);
        return;
    }
    *(slash_pos+1) = '\0';
    free(state.current_dir);
    state.current_dir = parent_dir;
}

/* Set the pattern run speed, by some means other than the speed slider. This function will
 * readjust the slider, thus triggering an event, so handle_speed_slider_change will take care of
 * everything else.
 */
void set_speed(int32 new_speed)
{
    GtkAdjustment*  slider_state;

    state.speed = new_speed;
    if (state.speed < 1)
        state.speed = 1;

    slider_state = gtk_range_get_adjustment(GTK_RANGE(gui.speed_slider));
    if (state.speed < slider_state->lower) {
        slider_state->lower = state.speed;
        gtk_adjustment_changed(slider_state);
    } else if (state.speed > slider_state->upper) {
        slider_state->upper = state.speed;
        gtk_adjustment_changed(slider_state);
        set_speed_label_size(state.speed);
    }
    gtk_adjustment_set_value(slider_state, state.speed);
}

/* Deselect any selected item in the sidebar.
 */
void sidebar_unselect(void)
{
    gtk_clist_unselect_all(GTK_CLIST(gui.pattern_list));
    GTK_CLIST(gui.pattern_list)->focus_row = -1;
    gtk_widget_queue_draw(gui.pattern_list);
}

/* Toggle state.pattern_running, and update menus, toolbar, etc. accordingly, but do *not* update
 * the canvas or the tick label.
 */
void start_stop(void)
{
    const char*  new_text;
    GtkWidget*   new_icon;

    state.pattern_running = !state.pattern_running;
    if (state.pattern_running) {
        state.skipped_frames = 0;
        state.start_tick = tick;
        state.start_time = get_time_milliseconds();
        new_text = "Stop";
        new_icon = gui.stop_pixmap;
    } else {
        state.skipped_frames = 0;
        new_text = "Start";
        new_icon = gui.start_pixmap;
    }

    gtk_label_set_text(GTK_LABEL(gui.start_stop_menu_label), new_text);
    gtk_container_add(GTK_CONTAINER(gui.start_stop_toolbar_button), new_icon);
    gtk_widget_show(new_icon);
}

/*** Utility Functions ***/

/* Blend the pixel rgb2 (an array of 3 unsigned chars represented red, green,
 * blue) onto the pixel rgb1, where rgb2 has the given opacity (0-255).
 */
void alpha_blend(uint8* rgb1, const uint8* rgb2, int32 opacity)
{
    *rgb1 = ((int32)(*rgb1++) * (255 - opacity) + (int32)(*rgb2++) * opacity) / 255;
    *rgb1 = ((int32)(*rgb1++) * (255 - opacity) + (int32)(*rgb2++) * opacity) / 255;
    *rgb1 = ((int32)(*rgb1++) * (255 - opacity) + (int32)(*rgb2++) * opacity) / 255;
}

/* Return a dimension structure representing the width and height of the given rectangle.
 */
void get_rect_dimensions(const rect* r, dimension* dim)
{
    dim->width  = r->end.x - r->start.x + 1;
    dim->height = r->end.y - r->start.y + 1;
}

/* Get the current time in milliseconds, as a 64-bit uint
 */
uint64 get_time_milliseconds(void)
{
    struct timeval   t;
    struct timezone  tz;

    gettimeofday(&t, &tz);
    return (uint64)(t.tv_sec) * 1000ULL + (uint64)(t.tv_usec) / 1000ULL;
}

/* Return true if the given string is composed entirely of numeric digits
 */
boolean is_numeric(const char* str)
{
    const char*  p;

    for (p=str; *p; p++) {
        if (!isdigit(*p))
            return FALSE;
    }
    return TRUE;
}

/* Load a GTK pixmap from an xpm file, returning NULL on failure.
 */
GtkWidget* load_pixmap_from_xpm_file(const char* path)
{
    GdkPixmap*  gdk_pixmap;

    gdk_pixmap = gdk_pixmap_create_from_xpm(gui.window->window, NULL,
                                            &(gui.window->style->bg[GTK_STATE_NORMAL]), path);
    if (!gdk_pixmap)
        return NULL;
    return gtk_pixmap_new(gdk_pixmap, NULL);
}

/* Load a GTK Pixmap from an array of RGBA pixel values, alpha blending onto the given
 * background color.
 */
GtkWidget* load_pixmap_from_rgba_array(uint8* pixels, int32 width, int32 height, GdkColor* bg)
{
    GdkPixmap*  gdk_pixmap;
    uint8       bg_rgb[3];
    uint8*      rgb_pixels;
    int32       i;

    bg_rgb[0] = bg->red   / 257;
    bg_rgb[1] = bg->green / 257;
    bg_rgb[2] = bg->blue  / 257;
    rgb_pixels = safe_malloc(width * height * 3);
    for (i=0; i < width*height; i++) {
        memcpy(&rgb_pixels[i*3], bg_rgb, 3);
        alpha_blend(&rgb_pixels[i*3], &pixels[i*4], pixels[i*4+3]);
    }
    gdk_pixmap = gdk_pixmap_new(gui.window->window, width, height, -1);
    gdk_draw_rgb_image(gdk_pixmap, gui.window->style->fg_gc[GTK_WIDGET_STATE(gui.window)], 0, 0,
                       width, height, GDK_RGB_DITHER_NONE, rgb_pixels, width*3);
    free(rgb_pixels);
    return gtk_pixmap_new(gdk_pixmap, NULL);
}

/* Insure that r->start is the upper-left-hand corner of the rectangle.
 */
void normalize_rectangle(rect* r)
{
    int32  temp;

    if (r->start.x > r->end.x)
        SWAP(r->start.x, r->end.x);
    if (r->start.y > r->end.y)
        SWAP(r->start.y, r->end.y);
}

/* Return TRUE if the two point structures are identical.
 */
boolean points_identical(const point* p1, const point* p2)
{
    return (p1->x == p2->x && p1->y == p2->y);
}

boolean rects_identical(const rect* p1, const rect* p2)
{
    return (points_identical(&(p1->start), &(p2->start)) &&
            points_identical(&(p1->end), &(p2->end)));
}
